/** --------------------------------------------------------------
 * IrrlichtΉ
 * author:         IrrLesson
 * date:           2008/04/27
 * email:          lesson@live.jp
 * web:            http://lesson.ifdef.jp/
 * --------------------------------------------------------------
 */

//
//                 Half-Life Model Viewer (c) 1999 by Mete Ciragan
//
// file:           ControlPanel.h
// last modified:  Oct 20 programs and associated files contained in this
//                 distribution were developed by Mete Ciragan. The programs
//                 are not in the public domain, but they are freely
//                 distributable without licensing fees. These programs are
//                 provided without guarantee or warrantee expressed or
//                 implied.
//
// version:        1.24
//
// email:          mete@swissquake.ch
// web:            http://www.swissquake.ch/chumbalum-soft/
//

#ifndef INCLUDED_CONTROLPANEL
#define INCLUDED_CONTROLPANEL

#include "StudioModel.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define IDC_TAB						1901
#define IDC_RENDERMODE				2001
#define IDC_TRANSPARENCY			2002
#define IDC_GROUND					2003
#define IDC_MIRROR					2004
#define IDC_BACKGROUND				2005
#define IDC_HITBOXES				2006
#define IDC_BONES					2007
#define IDC_ATTACHMENTS				2008

#define IDC_SEQUENCE				3001
#define IDC_SPEEDSCALE				3002
#define IDC_STOP					3003
#define IDC_PREVFRAME				3004
#define IDC_FRAME					3005
#define IDC_NEXTFRAME				3006

#define IDC_BODYPART				4001
#define IDC_SUBMODEL				4002
#define IDC_CONTROLLER				4003
#define IDC_CONTROLLERVALUE			4004
#define IDC_SKINS					4005

#define IDC_TEXTURES				5001
#define IDC_EXPORTTEXTURE			5002
#define IDC_IMPORTTEXTURE			5003
#define IDC_SAVEMODEL				5004
#define IDC_TEXTURESCALE			5005
#define IDC_CHROME					5006

#define IDC_FULLSCREEN				6001

class StudioModel;

class ControlPanel
{
	
public:

	StudioModel* m_studioModel;

	/**
	 * RXgN^
	 */
	ControlPanel()
	{
		InitViewerSettings ();
	}

	/**
	 * fXgN^
	 */
	~ControlPanel()
	{
	}

	/**
	 * r[Ȁ
	 */
	void InitViewerSettings ()
	{
		memset (&g_viewerSettings, 0, sizeof (ViewerSettings));
		g_viewerSettings.rot[0]       = -90.0f;
		g_viewerSettings.trans[3]     = 50.0f;
		g_viewerSettings.renderMode   = RM_TEXTURED;
		g_viewerSettings.transparency = 1.0f;

		g_viewerSettings.gColor[0]    = 0.85f;
		g_viewerSettings.gColor[1]    = 0.85f;
		g_viewerSettings.gColor[2]    = 0.69f;

		g_viewerSettings.lColor[0]    = 1.0f;
		g_viewerSettings.lColor[1]    = 1.0f;
		g_viewerSettings.lColor[2]    = 1.0f;

		g_viewerSettings.speedScale   = 1.0f;
		g_viewerSettings.textureLimit = 256;
	}

	/**
	 * fǂݍ
	 */
	void loadModel (const char *filename)
	{
		//Õf
		m_studioModel->FreeModel ();

		//wb_ǂݍ݁EeNX`o^
		if (m_studioModel->LoadModel ((char *) filename))
		{
			if (m_studioModel->PostLoadModel ((char *) filename))
			{
				initSequences ();
				initBodyparts ();
				initBoneControllers ();
				initSkins ();
				initTextures ();
				centerView ();
				strcpy (g_viewerSettings.modelFile, filename);
				setModelInfo ();
				g_viewerSettings.sequence = 0;
				g_viewerSettings.speedScale = 1.0f;
				//slSpeedScale->setValue (40);
				int i;
				for (i = 0; i < 32; i++)
					g_viewerSettings.submodels[i] = 0;
				for (i = 0; i < 8; i++)
					g_viewerSettings.controllers[i] = 0;
			}
		}
	}

	/**
	 * `惂[hݒ
	 */
	void setRenderMode (int mode)
	{
		g_viewerSettings.renderMode = mode;
	}

	/**
	 * nʕ`
	 */
	void setShowGround (bool b)
	{
		g_viewerSettings.showGround = b;
	}

	/**
	 * ~[
	 */
	void setMirror (bool b)
	{
		g_viewerSettings.useStencil = (!g_viewerSettings.use3dfx && b);
		g_viewerSettings.mirror = b;
	}

	/**
	 * wi`
	 */
	void setShowBackground (bool b)
	{
		g_viewerSettings.showBackground = b;
	}

	/**
	 * Aj[V̏
	 */
	void initSequences ()
	{
		studiohdr_t *hdr = m_studioModel->getStudioHeader ();
		if (hdr)
		{
	//		cSequence->removeAll ();
			for (int i = 0; i < hdr->numseq; i++)
			{
				mstudioseqdesc_t *pseqdescs = (mstudioseqdesc_t *) ((byte *) hdr + hdr->seqindex);
	//			cSequence->add (pseqdescs[i].label);
			}

	//		cSequence->select (0);
		}
	}

	/**
	 * Aj[Vݒ
	 */
	void setSequence (int index)
	{
	//	cSequence->select (index);
		m_studioModel->SetSequence(index);
		g_viewerSettings.sequence = index;
	}

	/**
	 * {fBp[c
	 */
	void initBodyparts ()
	{
		studiohdr_t *hdr = m_studioModel->getStudioHeader ();
		if (hdr)
		{
			int i;
			mstudiobodyparts_t *pbodyparts = (mstudiobodyparts_t *) ((byte *) hdr + hdr->bodypartindex);

	//		cBodypart->removeAll ();
			if (hdr->numbodyparts > 0)
			{
	//			for (i = 0; i < hdr->numbodyparts; i++)
	//				cBodypart->add (pbodyparts[i].name);

	//			cBodypart->select (0);

	//			cSubmodel->removeAll ();
				for (i = 0; i < pbodyparts[0].nummodels; i++)
				{
					char str[64];
					sprintf (str, "Submodel %d", i + 1);
	//				cSubmodel->add (str);
				}
	//			cSubmodel->select (0);
			}
		}
	}


	/**
	 * {fB[p[cݒigpj
	 */
	void setBodypart (int index)
	{
		studiohdr_t *hdr = m_studioModel->getStudioHeader ();
		if (hdr)
		{
			//cBodypart->setEn
	//		cBodypart->select (index);
			if (index < hdr->numbodyparts)
			{
				mstudiobodyparts_t *pbodyparts = (mstudiobodyparts_t *) ((byte *) hdr + hdr->bodypartindex);
	//			cSubmodel->removeAll ();
			
				for (int i = 0; i < pbodyparts[index].nummodels; i++)
				{
					char str[64];
					sprintf (str, "Submodel %d", i + 1);
	//				cSubmodel->add (str);
				}
	//			cSubmodel->select (0);
				//m_studioModel->SetBodygroup (index, 0);
			}
		}
	}
	
	/**
	 * Tufݒigpj
	 */
	void setSubmodel (int index)
	{
	//	m_studioModel->SetBodygroup (cBodypart->getSelectedIndex (), index);
	//	g_viewerSettings.submodels[cBodypart->getSelectedIndex ()] = index;
	}
	
	/**
	 * {[Rg[
	 */
	void initBoneControllers ()
	{
		studiohdr_t *hdr = m_studioModel->getStudioHeader ();
		if (hdr)
		{
	//		cController->setEnabled (hdr->numbonecontrollers > 0);
	//		slController->setEnabled (hdr->numbonecontrollers > 0);
	//		cController->removeAll ();

			mstudiobonecontroller_t *pbonecontrollers = (mstudiobonecontroller_t *) ((byte *) hdr + hdr->bonecontrollerindex);
			for (int i = 0; i < hdr->numbonecontrollers; i++)
			{
				char str[32];
				if (pbonecontrollers[i].index == 4)
					sprintf (str, "Mouth");
				else
					sprintf (str, "Controller %d", pbonecontrollers[i].index);
	//			cController->add (str);
			}

			if (hdr->numbonecontrollers > 0)
			{
	//			cController->select (0);
	//			slController->setRange ((int) pbonecontrollers[0].start, (int) pbonecontrollers[0].end);
	//			slController->setValue (0);
			}

		}
	}

	/**
	 * {[Rg[ݒigpj
	 */
	void setBoneController (int index)
	{
		studiohdr_t *hdr = m_studioModel->getStudioHeader ();
		if (hdr)
		{
			mstudiobonecontroller_t *pbonecontrollers = (mstudiobonecontroller_t *) ((byte *) hdr + hdr->bonecontrollerindex);
	//		slController->setRange ((int) pbonecontrollers[index].start, (int) pbonecontrollers[index].end);
	//		slController->setValue (0);
		}
	}

	/**
	 * {[Rg[lݒigpj
	 */
	void setBoneControllerValue (int index, float value)
	{
		studiohdr_t *hdr = m_studioModel->getStudioHeader ();
		if (hdr)
		{
			mstudiobonecontroller_t *pbonecontrollers = (mstudiobonecontroller_t *) ((byte *) hdr + hdr->bonecontrollerindex);
			if (pbonecontrollers[index].index == 4)
				m_studioModel->SetMouth (value);
			else
				m_studioModel->SetController (pbonecontrollers[index].index, value);
		
			g_viewerSettings.controllers[index] = value;
		}
	}

	/**
	 * XL
	 */
	void initSkins ()
	{
		studiohdr_t *hdr = m_studioModel->getStudioHeader ();
		if (hdr)
		{
	//		cSkin->setEnabled (hdr->numskinfamilies > 0);
	//		cSkin->removeAll ();

			for (int i = 0; i < hdr->numskinfamilies; i++)
			{
				char str[32];
				sprintf (str, "Skin %d", i + 1);
	//			cSkin->add (str);
			}

	//		cSkin->select (0);
			m_studioModel->SetSkin (0);
			g_viewerSettings.skin = 0;
		}
	}

	/**
	 * fݒ
	 */
	void setModelInfo ()
	{
		static char str[2048];
		studiohdr_t *hdr = m_studioModel->getStudioHeader ();

		if (!hdr)
			return;

		sprintf (str,
			"Bones: %d\n"
			"Bone Controllers: %d\n"
			"Hit Boxes: %d\n"
			"Sequences: %d\n"
			"Sequence Groups: %d\n",
			hdr->numbones,
			hdr->numbonecontrollers,
			hdr->numhitboxes,
			hdr->numseq,
			hdr->numseqgroups
			);

	//	lModelInfo1->setLabel (str);

		sprintf (str,
			"Textures: %d\n"
			"Skin Families: %d\n"
			"Bodyparts: %d\n"
			"Attachments: %d\n"
			"Transitions: %d\n",
			hdr->numtextures,
			hdr->numskinfamilies,
			hdr->numbodyparts,
			hdr->numattachments,
			hdr->numtransitions);

	//	lModelInfo2->setLabel (str);
	}

	/**
	 * eNX`
	 */
	void initTextures ()
	{
		studiohdr_t *hdr = m_studioModel->getTextureHeader ();
		if (hdr)
		{
	//		cTextures->removeAll ();
			mstudiotexture_t *ptextures = (mstudiotexture_t *) ((byte *) hdr + hdr->textureindex);
	//		for (int i = 0; i < hdr->numtextures; i++)
	//			cTextures->add (ptextures[i].name);
	//		cTextures->select (0);
			g_viewerSettings.texture = 0;
	//		if (hdr->numtextures > 0)
	//			cbChrome->setChecked ((ptextures[0].flags & STUDIO_NF_CHROME) == STUDIO_NF_CHROME);
		}
	}

	/**
	 * \ʒu
	 */
	void centerView ()
	{
		float min[3], max[3];
		m_studioModel->ExtractBbox (min, max);

		float dx = max[0] - min[0];
		float dy = max[1] - min[1];
		float dz = max[2] - min[2];
		float d = dx;
		if (dy > d)
			d = dy;
		if (dz > d)
			d = dz;
		g_viewerSettings.trans[0] = 0;
		g_viewerSettings.trans[1] = min[2] + dz / 2;
		g_viewerSettings.trans[2] = d * 1.0f;
		g_viewerSettings.rot[0] = -90.0f;
		g_viewerSettings.rot[1] = -90.0f;
		g_viewerSettings.rot[2] = 0.0f;
	}

};


#endif // INCLUDED_CONTROLPANEL