/** --------------------------------------------------------------
 * IrrlichtΉ
 * author:         IrrLesson
 * date:           2008/04/27
 * email:          lesson@live.jp
 * web:            http://lesson.ifdef.jp/
 * --------------------------------------------------------------
 */

/***
*
*	Copyright (c) 1998, Valve LLC. All rights reserved.
*	
*	This product contains software technology licensed from Id 
*	Software, Inc. ("Id Technology").  Id Technology (c) 1996 Id Software, Inc. 
*	All Rights Reserved.
*
****/

#ifndef INCLUDED_STUDIOMODEL
#define INCLUDED_STUDIOMODEL

#include "mdlviewer.h"
#include "studio.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <irrlicht.h>

using namespace irr;
using namespace core;
using namespace video;
using namespace scene;

enum // render modes
{
	RM_WIREFRAME,
	RM_FLATSHADED,
	RM_SMOOTHSHADED,
	RM_TEXTURED
};

typedef struct
{
	// model 
	float rot[3];
	float trans[3];

	// render
	int renderMode;
	float transparency;
	bool showBackground;
	bool showGround;
	bool showHitBoxes;
	bool showBones;
	bool showTexture;
	bool showAttachments;
	int texture;
	float textureScale;
	int skin;
	bool mirror;
	bool useStencil;	// if 3dfx fullscreen set false

	// animation
	int sequence;
	float speedScale;

	// bodyparts and bonecontrollers
	int submodels[32];
	float controllers[8];

	// fullscreen
	int width, height;
	bool use3dfx;
	bool cds;

	// colors
	float bgColor[4];
	float lColor[4];
	float gColor[4];

	// misc
	int textureLimit;
	bool pause;

	// only used for fullscreen mode
	char modelFile[256];
	char backgroundTexFile[256];
	char groundTexFile[256];
} ViewerSettings;

ViewerSettings  g_viewerSettings;


class StudioModel
{
	IVideoDriver* m_driver;
	ITexture *m_texture[16];

	#define TEX_NUM 0//3
	#define MAX_TEXTURE_DIMS 512

	int             m_texnum;
	vec3_t			m_xformverts[MAXSTUDIOVERTS];	// transformed vertices
	vec3_t			m_lightvalues[MAXSTUDIOVERTS];	// light surface normals
	vec3_t			*m_pxformverts;
	vec3_t			*m_pvlightvalues;
	vec3_t			m_lightvec;						// light vector in model reference frame
	vec3_t			m_blightvec[MAXSTUDIOBONES];	// light vectors in bone reference frames
	int				m_ambientlight;					// ambient world light
	float			m_shadelight;					// direct world light
	vec3_t			m_lightcolor;
	int				m_smodels_total;				// cookie
	float			m_bonetransform[MAXSTUDIOBONES][3][4];	// bone transformation matrix
	int				m_chrome[MAXSTUDIOVERTS][2];	// texture coords for surface normals
	int				m_chromeage[MAXSTUDIOBONES];	// last time chrome vectors were updated
	vec3_t			m_chromeup[MAXSTUDIOBONES];		// chrome vector "up" in bone reference frames
	vec3_t			m_chromeright[MAXSTUDIOBONES];	// chrome vector "right" in bone reference frames



public:
	/**
	 * RXgN^
	 */
	StudioModel(IVideoDriver* driver) : m_driver(driver)
	{
		memset(m_origin,0x00,sizeof(m_origin));
		memset(m_angles,0x00,sizeof(m_angles));
		m_sequence    = 0;
		m_frame       = 0;
		m_bodynum     = 0;
		m_skinnum     = 0;
		memset(m_controller,0x00,sizeof(m_controller));
		memset(m_blending,0x00,sizeof(m_blending));
		m_mouth       = 0;
		m_owntexmodel = 0;

		
		m_pstudiohdr  = 0;
		m_pmodel      = 0;

		m_ptexturehdr = 0;
		memset(m_panimhdr, 0x00, sizeof(m_panimhdr));

		memset(m_adj,0x00,sizeof(m_adj));

		m_texnum = TEX_NUM;
	}

	/**
	 * fXgN^
	 */
	~StudioModel()
	{
	}

	/**
	 * X^WIwb_擾
	 */
	studiohdr_t *getStudioHeader () const
	{
		return m_pstudiohdr;
	}

	/**
	 * eNX`wb_擾
	 */
	studiohdr_t *getTextureHeader () const 
	{
		return m_ptexturehdr;
	}

	/**
	 * Aj[Vwb_擾
	 */
	studiohdr_t *getAnimHeader (int i) const 
	{
		return m_panimhdr[i];
	}

	/**
	 * eNX`XV
	 */
	void UploadTexture( mstudiotexture_t *ptexture, byte *data, byte *pal, int name )
	{
		//`FbN
		if(ptexture == NULL || data == NULL || pal == NULL)
			return;

		//摜̍╝͍ő512ƂFMAX_TEXTURE_DIMS=512
		int	row1[MAX_TEXTURE_DIMS];
		int row2[MAX_TEXTURE_DIMS];
		int col1[MAX_TEXTURE_DIMS];
		int col2[MAX_TEXTURE_DIMS];
		int outWidth, outHeight;
		byte* tex;
		byte* out;
		byte *pix1, *pix2, *pix3, *pix4;
		int x,y;
		IImage *image;

		//摜TCY2nɍ킹
		for(outWidth = 1; outWidth < ptexture->width; outWidth <<= 1);
		if(outWidth > MAX_TEXTURE_DIMS)
			outWidth = MAX_TEXTURE_DIMS;
		for(outHeight = 1; outHeight < ptexture->height; outHeight <<= 1);
		if(outHeight > MAX_TEXTURE_DIMS)
			outHeight = MAX_TEXTURE_DIMS;

		//eNX`pɃ蓖
		tex = (u8*)malloc(outWidth*outHeight*4);
		out = tex;
		if(out == NULL)
			return;

		//̊
		for(x=0;x<outWidth;x++)
		{
			col1[x] = (s32)((x + 0.25) * (ptexture->width / (f32)outWidth));
			col2[x] = (s32)((x + 0.75) * (ptexture->width / (f32)outWidth));
		}

		//c̊
		for(y=0;y<outHeight;y++)
		{
			row1[y] = (s32)((y + 0.25) * (ptexture->height / (f32)outHeight)) * ptexture->width;
			row2[y] = (s32)((y + 0.75) * (ptexture->height / (f32)outHeight)) * ptexture->width;
		}

		//eNX`f[^̍쐬
		for(y=0;y<outHeight;y++)
		{
			for(x=0;x<outWidth;x++,out+=4)
			{
				pix1 = &pal[data[row1[y] + col1[x]] * 3];
				pix2 = &pal[data[row1[y] + col2[x]] * 3];
				pix3 = &pal[data[row2[y] + col1[x]] * 3];
				pix4 = &pal[data[row2[y] + col2[x]] * 3];

				//RGB̏ԂtȂ̂Œ
				out[0] = (pix1[2] + pix2[2] + pix3[2] + pix4[2])>>2;//B
				out[1] = (pix1[1] + pix2[1] + pix3[1] + pix4[1])>>2;//G
				out[2] = (pix1[0] + pix2[0] + pix3[0] + pix4[0])>>2;//R
				out[3] = 0xFF;										//A
			}
		}

		//C[W̍쐬
		image = m_driver->createImageFromData(
			ECF_A8R8G8B8, dimension2d<s32>(outWidth, outHeight), tex);

		//eNX`o^
		m_texture[name-TEX_NUM] = 
			m_driver->addTexture(ptexture->name, image);

		//eNX`J
		image->drop();//K{
		free(tex);
		tex = NULL;
	}

	/**
	 * f
	 */
	void FreeModel ()
	{
		if (m_pstudiohdr)
			free (m_pstudiohdr);

		if (m_ptexturehdr && m_owntexmodel)
			free (m_ptexturehdr);

		m_pstudiohdr = m_ptexturehdr = 0;
		m_owntexmodel = false;

		int i;
		for (i = 0; i < 32; i++)
		{
			if (m_panimhdr[i])
			{
				//mallocŊmۂeNX`pJ
				free (m_panimhdr[i]);
				m_panimhdr[i] = 0;
			}
		}

		//eNX`J
		for(i=0;i<m_texnum;i++)
		{
			m_driver->removeTexture(m_texture[i]);
			m_texture[i] = NULL;
		}
		m_driver->removeAllTextures();

		m_texnum = TEX_NUM;
	}

	/**
	 * fǂݍ
	 */
	studiohdr_t *LoadModel( char *modelname )
	{
		FILE *fp;
		long size;
		void *buffer;

		if (!modelname)
			return 0;

		// load the model
		if( (fp = fopen( modelname, "rb" )) == NULL)
			return 0;

		fseek( fp, 0, SEEK_END );
		size = ftell( fp );
		fseek( fp, 0, SEEK_SET );

		//fǂݍݗpeNX`m
		buffer = malloc( size );
		if (!buffer)
		{
			fclose (fp);
			return 0;
		}

		//ff[^ǂݍ
		fread( buffer, size, 1, fp );
		fclose( fp );

		byte				*pin;
		studiohdr_t			*phdr;
		mstudiotexture_t	*ptexture;

		pin      = (byte *)buffer;
		phdr     = (studiohdr_t *)pin;
		ptexture = (mstudiotexture_t *)(pin + phdr->textureindex);

		//wb_`FbN
		if (strncmp ((const char *) buffer, "IDST", 4) &&
			strncmp ((const char *) buffer, "IDSQ", 4))
		{
			free (buffer);
			return 0;
		}

		if (!strncmp ((const char *) buffer, "IDSQ", 4) && !m_pstudiohdr)
		{
			free (buffer);
			return 0;
		}

		//eNX``FbN
		if (phdr->textureindex > 0 && phdr->numtextures <= MAXSTUDIOSKINS)
		{
			int n = phdr->numtextures;
			for (int i = 0; i < n; i++)
			{
				// strcpy( name, mod->name );
				// strcpy( name, ptexture[i].name );
				//eNX`ǂݍ
				UploadTexture( &ptexture[i], pin + ptexture[i].index, pin + ptexture[i].width * ptexture[i].height + ptexture[i].index, m_texnum++ );
			}
		}

		// UNDONE: free texture memory
		//ff[^̃|C^Zbg
		if (!m_pstudiohdr)
			m_pstudiohdr = (studiohdr_t *)buffer;

		return (studiohdr_t *)buffer;
	}

	/**
	 * fǂݍ
	 */
	bool PostLoadModel ( char *modelname )
	{
		if(m_pstudiohdr ==NULL)
			return false;

		// preload textures
		if (m_pstudiohdr->numtextures == 0)
		{
			char texturename[256];

			strcpy( texturename, modelname );
			strcpy( &texturename[strlen(texturename) - 4], "T.mdl" );

			m_ptexturehdr = LoadModel( texturename );
			if (!m_ptexturehdr)
			{
				FreeModel ();
				return false;
			}
			m_owntexmodel = true;
		}
		else
		{
			m_ptexturehdr = m_pstudiohdr;
			m_owntexmodel = false;
		}

		// preload animations
		//Aj[V̓ǂݍ
		if (m_pstudiohdr->numseqgroups > 1)
		{
			for (int i = 1; i < m_pstudiohdr->numseqgroups; i++)
			{
				char seqgroupname[256];

				strcpy( seqgroupname, modelname );
				sprintf( &seqgroupname[strlen(seqgroupname) - 4], "%02d.mdl", i );

				m_panimhdr[i] = LoadModel( seqgroupname );
				if (!m_panimhdr[i])
				{
					FreeModel ();
					return false;
				}
			}
		}

		SetSequence (0);
		SetController (0, 0.0f);
		SetController (1, 0.0f);
		SetController (2, 0.0f);
		SetController (3, 0.0f);
		SetMouth (0.0f);

		//{fB̐ݒ
		int n;
		for (n = 0; n < m_pstudiohdr->numbodyparts; n++)
			SetBodygroup (n, 0);

		SetSkin (0);
	/*
		vec3_t mins, maxs;
		ExtractBbox (mins, maxs);
		if (mins[2] < 5.0f)
			m_origin[2] = -mins[2];
	*/
		return true;
	}

	/**
	 * fۑigpj
	 */
	bool SaveModel ( char *modelname )
	{
		if (!modelname)
			return false;

		if (!m_pstudiohdr)
			return false;

		FILE *file;
		
		file = fopen (modelname, "wb");
		if (!file)
			return false;

		fwrite (m_pstudiohdr, sizeof (byte), m_pstudiohdr->length, file);
		fclose (file);

		// write texture model
		if (m_owntexmodel && m_ptexturehdr)
		{
			char texturename[256];

			strcpy( texturename, modelname );
			strcpy( &texturename[strlen(texturename) - 4], "T.mdl" );

			file = fopen (texturename, "wb");
			if (file)
			{
				fwrite (m_ptexturehdr, sizeof (byte), m_ptexturehdr->length, file);
				fclose (file);
			}
		}

		// write seq groups
		if (m_pstudiohdr->numseqgroups > 1)
		{
			for (int i = 1; i < m_pstudiohdr->numseqgroups; i++)
			{
				char seqgroupname[256];

				strcpy( seqgroupname, modelname );
				sprintf( &seqgroupname[strlen(seqgroupname) - 4], "%02d.mdl", i );

				file = fopen (seqgroupname, "wb");
				if (file)
				{
					fwrite (m_panimhdr[i], sizeof (byte), m_panimhdr[i]->length, file);
					fclose (file);
				}
			}
		}
		return true;
	}

	/**
	 * f`
	 */
	void DrawModel( void )
	{
		int i;

		if (!m_pstudiohdr)
			return;

		m_smodels_total++; // render data cache cookie

		m_pxformverts   = &m_xformverts[0];
		m_pvlightvalues = &m_lightvalues[0];

		if (m_pstudiohdr->numbodyparts == 0)
			return;

		SetUpBones ( );//{[ݒ

		SetupLighting( );//Cgݒ

		//`Jn
		for (i=0 ; i < m_pstudiohdr->numbodyparts ; i++) 
		{
			SetupModel( i );
			if (g_viewerSettings.transparency > 0.0f)
				//`{
				DrawPoints( );
		}

		//{[`
		if (g_viewerSettings.showBones && !g_viewerSettings.use3dfx)
		{
			mstudiobone_t *pbones = (mstudiobone_t *) ((byte *) m_pstudiohdr + m_pstudiohdr->boneindex);

			for (i = 0; i < m_pstudiohdr->numbones; i++)
			{
				if (pbones[i].parent >= 0)
				{
					SColor color = 0xFFFFFF00;

					//ݒ
					SMaterial Material;
					Material.setFlag(EMF_LIGHTING, false);	//Cgݒ
					Material.GouraudShading = true;
					Material.ZBuffer = 0;
					m_driver->setMaterial(Material); 

					//n_iԁj
					m_driver->draw3DBox(
						aabbox3d<f32>(
							m_bonetransform[pbones[i].parent][0][3], 
							m_bonetransform[pbones[i].parent][2][3], 
							m_bonetransform[pbones[i].parent][1][3],
							m_bonetransform[pbones[i].parent][0][3]+0.1, 
							m_bonetransform[pbones[i].parent][2][3]+0.1, 
							m_bonetransform[pbones[i].parent][1][3]+0.1
						), 0xFFFF0000);

					m_driver->draw3DLine(
						vector3df(
							m_bonetransform[pbones[i].parent][0][3], 
							m_bonetransform[pbones[i].parent][2][3], 
							m_bonetransform[pbones[i].parent][1][3]), 
						vector3df(
							m_bonetransform[i][0][3], 
							m_bonetransform[i][2][3], 
							m_bonetransform[i][1][3]), 
						color);

					//ԓ_ij
					if (pbones[pbones[i].parent].parent != -1)
						m_driver->draw3DBox(
							aabbox3d<f32>(
								m_bonetransform[pbones[i].parent][0][3], 
								m_bonetransform[pbones[i].parent][2][3], 
								m_bonetransform[pbones[i].parent][1][3],
								m_bonetransform[pbones[i].parent][0][3]+0.1, 
								m_bonetransform[pbones[i].parent][2][3]+0.1, 
								m_bonetransform[pbones[i].parent][1][3]+0.1
							), 0xFF0000FF);

					//I_ij
					m_driver->draw3DBox(
						aabbox3d<f32>(
							m_bonetransform[i][0][3], 
							m_bonetransform[i][2][3], 
							m_bonetransform[i][1][3],
							m_bonetransform[i][0][3]+0.1, 
							m_bonetransform[i][2][3]+0.1, 
							m_bonetransform[i][1][3]+0.1
						), 0xFF0000FF);
				}
				else
				{
					// draw parent bone node
	//				glPointSize (5.0f);
	//				glColor3f (0.8f, 0, 0);
	//				glBegin (GL_POINTS);
	//				glVertex3f (m_bonetransform[i][0][3], m_bonetransform[i][1][3], m_bonetransform[i][2][3]);
	//				glEnd ();
				}
			}

	//		glPointSize (1.0f);
		}

		//A^b`gij
		if (g_viewerSettings.showAttachments)// && !g_viewerSettings.use3dfx)
		{
			for (i = 0; i < m_pstudiohdr->numattachments; i++)
			{
				mstudioattachment_t *pattachments = (mstudioattachment_t *) ((byte *) m_pstudiohdr + m_pstudiohdr->attachmentindex);
				vec3_t v[4];
				VectorTransform (pattachments[i].org, m_bonetransform[pattachments[i].bone], v[0]);
				VectorTransform (pattachments[i].vectors[0], m_bonetransform[pattachments[i].bone], v[1]);
				VectorTransform (pattachments[i].vectors[1], m_bonetransform[pattachments[i].bone], v[2]);
				VectorTransform (pattachments[i].vectors[2], m_bonetransform[pattachments[i].bone], v[3]);


				SColor color = 0xFFFFFFFF;

				//ݒ
				SMaterial Material;
				Material.setFlag(EMF_LIGHTING, false);	//Cgݒ
				Material.GouraudShading = true;
				Material.ZBuffer = 0;
				m_driver->setMaterial(Material); 

				m_driver->draw3DLine(
					vector3df(v[0][0], v[0][2], v[0][1]), 
					vector3df(v[1][0], v[1][2], v[1][1]), 
					color);
				m_driver->draw3DLine(
					vector3df(v[0][0], v[0][2], v[0][1]), 
					vector3df(v[2][0], v[2][2], v[2][1]), 
					color);
				m_driver->draw3DLine(
					vector3df(v[0][0], v[0][2], v[0][1]), 
					vector3df(v[3][0], v[3][2], v[3][1]), 
					color);


				//I_i΁j
				m_driver->draw3DBox(
					aabbox3d<f32>(
						v[0][0], 
						v[0][2], 
						v[0][1],
						v[0][0]+0.1, 
						v[0][2]+0.1, 
						v[0][1]+0.1
					), 0xFF00FF00);
			}
		}

		//qbg{bNXiԐj
		if (g_viewerSettings.showHitBoxes)
		{
			for (i = 0; i < m_pstudiohdr->numhitboxes; i++)
			{
				mstudiobbox_t *pbboxes = (mstudiobbox_t *) ((byte *) m_pstudiohdr + m_pstudiohdr->hitboxindex);
				vec3_t v[8], v2[8], bbmin, bbmax;

				VectorCopy (pbboxes[i].bbmin, bbmin);
				VectorCopy (pbboxes[i].bbmax, bbmax);

				v[0][0] = bbmin[0];
				v[0][1] = bbmax[1];
				v[0][2] = bbmin[2];

				v[1][0] = bbmin[0];
				v[1][1] = bbmin[1];
				v[1][2] = bbmin[2];

				v[2][0] = bbmax[0];
				v[2][1] = bbmax[1];
				v[2][2] = bbmin[2];

				v[3][0] = bbmax[0];
				v[3][1] = bbmin[1];
				v[3][2] = bbmin[2];

				v[4][0] = bbmax[0];
				v[4][1] = bbmax[1];
				v[4][2] = bbmax[2];

				v[5][0] = bbmax[0];
				v[5][1] = bbmin[1];
				v[5][2] = bbmax[2];

				v[6][0] = bbmin[0];
				v[6][1] = bbmax[1];
				v[6][2] = bbmax[2];

				v[7][0] = bbmin[0];
				v[7][1] = bbmin[1];
				v[7][2] = bbmax[2];

				VectorTransform (v[0], m_bonetransform[pbboxes[i].bone], v2[0]);
				VectorTransform (v[1], m_bonetransform[pbboxes[i].bone], v2[1]);
				VectorTransform (v[2], m_bonetransform[pbboxes[i].bone], v2[2]);
				VectorTransform (v[3], m_bonetransform[pbboxes[i].bone], v2[3]);
				VectorTransform (v[4], m_bonetransform[pbboxes[i].bone], v2[4]);
				VectorTransform (v[5], m_bonetransform[pbboxes[i].bone], v2[5]);
				VectorTransform (v[6], m_bonetransform[pbboxes[i].bone], v2[6]);
				VectorTransform (v[7], m_bonetransform[pbboxes[i].bone], v2[7]);
				
				//qbg{bNX̕`
				drawBox (v2);
			}
		}
	}

	/**
	 * Aj[Vt[̒
	 */
	void AdvanceFrame( float dt )
	{
		if (!m_pstudiohdr)
			return;

		mstudioseqdesc_t	*pseqdesc;
		pseqdesc = (mstudioseqdesc_t *)((byte *)m_pstudiohdr + m_pstudiohdr->seqindex) + m_sequence;

		if (dt > 0.1)
			dt = 0.1f;
		m_frame += dt * pseqdesc->fps;

		if (pseqdesc->numframes <= 1)
		{
			m_frame = 0;
		}
		else
		{
			// wrap
			m_frame -= (int)(m_frame / (pseqdesc->numframes - 1)) * (pseqdesc->numframes - 1);
		}
	}

	/**
	 * Aj[Vt[w
	 */
	int  SetFrame (int nFrame)
	{
		if (nFrame == -1)
			return m_frame;

		if (!m_pstudiohdr)
			return 0;

		mstudioseqdesc_t	*pseqdesc;
		pseqdesc = (mstudioseqdesc_t *)((byte *)m_pstudiohdr + m_pstudiohdr->seqindex) + m_sequence;

		m_frame = nFrame;

		if (pseqdesc->numframes <= 1)
		{
			m_frame = 0;
		}
		else
		{
			// wrap
			m_frame -= (int)(m_frame / (pseqdesc->numframes - 1)) * (pseqdesc->numframes - 1);
		}

		return m_frame;
	}

	/**
	 * 
	 */
	void ExtractBbox( float *mins, float *maxs )
	{
		mstudioseqdesc_t	*pseqdesc;

		pseqdesc = (mstudioseqdesc_t *)((byte *)m_pstudiohdr + m_pstudiohdr->seqindex);
		
		mins[0] = pseqdesc[ m_sequence ].bbmin[0];
		mins[1] = pseqdesc[ m_sequence ].bbmin[1];
		mins[2] = pseqdesc[ m_sequence ].bbmin[2];

		maxs[0] = pseqdesc[ m_sequence ].bbmax[0];
		maxs[1] = pseqdesc[ m_sequence ].bbmax[1];
		maxs[2] = pseqdesc[ m_sequence ].bbmax[2];
	}

	/**
	 * Aj[Vԍ擾
	 */
	int GetSequence(  )
	{
		return m_sequence;
	}

	/**
	 * Aj[Vԍw
	 */
	int SetSequence( int iSequence )
	{
		if (iSequence > m_pstudiohdr->numseq)
			return m_sequence;

		m_sequence = iSequence;
		m_frame = 0;

		return m_sequence;
	}

	/**
	 * Aj[V擾igpj
	 */
	void GetSequenceInfo( float *pflFrameRate, float *pflGroundSpeed )
	{
		mstudioseqdesc_t	*pseqdesc;

		pseqdesc = (mstudioseqdesc_t *)((byte *)m_pstudiohdr + m_pstudiohdr->seqindex) + (int)m_sequence;

		if (pseqdesc->numframes > 1)
		{
			*pflFrameRate = 256 * pseqdesc->fps / (pseqdesc->numframes - 1);
			*pflGroundSpeed = sqrt( pseqdesc->linearmovement[0]*pseqdesc->linearmovement[0]+ pseqdesc->linearmovement[1]*pseqdesc->linearmovement[1]+ pseqdesc->linearmovement[2]*pseqdesc->linearmovement[2] );
			*pflGroundSpeed = *pflGroundSpeed * pseqdesc->fps / (pseqdesc->numframes - 1);
		}
		else
		{
			*pflFrameRate = 256.0;
			*pflGroundSpeed = 0.0;
		}
	}

	/**
	 * Rg[ݒ
	 */
	float SetController( int iController, float flValue )
	{
		if (!m_pstudiohdr)
			return 0.0f;

		mstudiobonecontroller_t	*pbonecontroller = (mstudiobonecontroller_t *)((byte *)m_pstudiohdr + m_pstudiohdr->bonecontrollerindex);

		// find first controller that matches the index
		int i;
		for (i = 0; i < m_pstudiohdr->numbonecontrollers; i++, pbonecontroller++)
		{
			if (pbonecontroller->index == iController)
				break;
		}
		if (i >= m_pstudiohdr->numbonecontrollers)
			return flValue;

		// wrap 0..360 if it's a rotational controller
		if (pbonecontroller->type & (STUDIO_XR | STUDIO_YR | STUDIO_ZR))
		{
			// ugly hack, invert value if end < start
			if (pbonecontroller->end < pbonecontroller->start)
				flValue = -flValue;

			// does the controller not wrap?
			if (pbonecontroller->start + 359.0 >= pbonecontroller->end)
			{
				if (flValue > ((pbonecontroller->start + pbonecontroller->end) / 2.0) + 180)
					flValue = flValue - 360;
				if (flValue < ((pbonecontroller->start + pbonecontroller->end) / 2.0) - 180)
					flValue = flValue + 360;
			}
			else
			{
				if (flValue > 360)
					flValue = flValue - (int)(flValue / 360.0) * 360.0;
				else if (flValue < 0)
					flValue = flValue + (int)((flValue / -360.0) + 1) * 360.0;
			}
		}

		int setting = (int) (255 * (flValue - pbonecontroller->start) /
		(pbonecontroller->end - pbonecontroller->start));

		if (setting < 0) setting = 0;
		if (setting > 255) setting = 255;
		m_controller[iController] = setting;

		return setting * (1.0 / 255.0) * (pbonecontroller->end - pbonecontroller->start) + pbonecontroller->start;
	}

	/**
	 * 
	 */
	float SetMouth( float flValue )
	{
		if (!m_pstudiohdr)
			return 0.0f;

		mstudiobonecontroller_t	*pbonecontroller = (mstudiobonecontroller_t *)((byte *)m_pstudiohdr + m_pstudiohdr->bonecontrollerindex);

		// find first controller that matches the mouth
		for (int i = 0; i < m_pstudiohdr->numbonecontrollers; i++, pbonecontroller++)
		{
			if (pbonecontroller->index == 4)
				break;
		}

		// wrap 0..360 if it's a rotational controller
		if (pbonecontroller->type & (STUDIO_XR | STUDIO_YR | STUDIO_ZR))
		{
			// ugly hack, invert value if end < start
			if (pbonecontroller->end < pbonecontroller->start)
				flValue = -flValue;

			// does the controller not wrap?
			if (pbonecontroller->start + 359.0 >= pbonecontroller->end)
			{
				if (flValue > ((pbonecontroller->start + pbonecontroller->end) / 2.0) + 180)
					flValue = flValue - 360;
				if (flValue < ((pbonecontroller->start + pbonecontroller->end) / 2.0) - 180)
					flValue = flValue + 360;
			}
			else
			{
				if (flValue > 360)
					flValue = flValue - (int)(flValue / 360.0) * 360.0;
				else if (flValue < 0)
					flValue = flValue + (int)((flValue / -360.0) + 1) * 360.0;
			}
		}

		int setting = (int) (64 * (flValue - pbonecontroller->start) / (pbonecontroller->end - pbonecontroller->start));

		if (setting < 0) setting = 0;
		if (setting > 64) setting = 64;
		m_mouth = setting;

		return setting * (1.0 / 64.0) * (pbonecontroller->end - pbonecontroller->start) + pbonecontroller->start;
	}

	/**
	 * ufBOݒ
	 */
	float SetBlending( int iBlender, float flValue )
	{
		mstudioseqdesc_t	*pseqdesc;

		if (!m_pstudiohdr)
			return 0.0f;

		pseqdesc = (mstudioseqdesc_t *)((byte *)m_pstudiohdr + m_pstudiohdr->seqindex) + (int)m_sequence;

		if (pseqdesc->blendtype[iBlender] == 0)
			return flValue;

		if (pseqdesc->blendtype[iBlender] & (STUDIO_XR | STUDIO_YR | STUDIO_ZR))
		{
			// ugly hack, invert value if end < start
			if (pseqdesc->blendend[iBlender] < pseqdesc->blendstart[iBlender])
				flValue = -flValue;

			// does the controller not wrap?
			if (pseqdesc->blendstart[iBlender] + 359.0 >= pseqdesc->blendend[iBlender])
			{
				if (flValue > ((pseqdesc->blendstart[iBlender] + pseqdesc->blendend[iBlender]) / 2.0) + 180)
					flValue = flValue - 360;
				if (flValue < ((pseqdesc->blendstart[iBlender] + pseqdesc->blendend[iBlender]) / 2.0) - 180)
					flValue = flValue + 360;
			}
		}

		int setting = (int) (255 * (flValue - pseqdesc->blendstart[iBlender]) / (pseqdesc->blendend[iBlender] - pseqdesc->blendstart[iBlender]));

		if (setting < 0) setting = 0;
		if (setting > 255) setting = 255;

		m_blending[iBlender] = setting;

		return setting * (1.0 / 255.0) * (pseqdesc->blendend[iBlender] - pseqdesc->blendstart[iBlender]) + pseqdesc->blendstart[iBlender];
	}

	/**
	 * {fBO[vݒ
	 */
	int SetBodygroup( int iGroup, int iValue )
	{
		if (!m_pstudiohdr)
			return 0;

		if (iGroup > m_pstudiohdr->numbodyparts)
			return -1;

		mstudiobodyparts_t *pbodypart = (mstudiobodyparts_t *)((byte *)m_pstudiohdr + m_pstudiohdr->bodypartindex) + iGroup;

		int iCurrent = (m_bodynum / pbodypart->base) % pbodypart->nummodels;

		if (iValue >= pbodypart->nummodels)
			return iCurrent;

		m_bodynum = (m_bodynum - (iCurrent * pbodypart->base) + (iValue * pbodypart->base));

		return iValue;
	}

	/**
	 * XLݒ
	 */
	int SetSkin( int iValue )
	{
		if (!m_pstudiohdr)
			return 0;

		if (iValue >= m_pstudiohdr->numskinfamilies)
		{
			return m_skinnum;
		}

		m_skinnum = iValue;

		return iValue;
	}

	/**
	 * bVXP[ݒ
	 */
	void scaleMeshes (float scale)
	{
		if (!m_pstudiohdr)
			return;

		int i, j, k;

		// scale verts
		int tmp = m_bodynum;
		for (i = 0; i < m_pstudiohdr->numbodyparts; i++)
		{
			mstudiobodyparts_t *pbodypart = (mstudiobodyparts_t *)((byte *)m_pstudiohdr + m_pstudiohdr->bodypartindex) + i;
			for (j = 0; j < pbodypart->nummodels; j++)
			{
				SetBodygroup (i, j);
				SetupModel (i);

				vec3_t *pstudioverts = (vec3_t *)((byte *)m_pstudiohdr + m_pmodel->vertindex);

				for (k = 0; k < m_pmodel->numverts; k++)
					VectorScale (pstudioverts[k], scale, pstudioverts[k]);
			}
		}

		m_bodynum = tmp;

		// scale complex hitboxes
		mstudiobbox_t *pbboxes = (mstudiobbox_t *) ((byte *) m_pstudiohdr + m_pstudiohdr->hitboxindex);
		for (i = 0; i < m_pstudiohdr->numhitboxes; i++)
		{
			VectorScale (pbboxes[i].bbmin, scale, pbboxes[i].bbmin);
			VectorScale (pbboxes[i].bbmax, scale, pbboxes[i].bbmax);
		}

		// scale bounding boxes
		mstudioseqdesc_t *pseqdesc = (mstudioseqdesc_t *)((byte *)m_pstudiohdr + m_pstudiohdr->seqindex);
		for (i = 0; i < m_pstudiohdr->numseq; i++)
		{
			VectorScale (pseqdesc[i].bbmin, scale, pseqdesc[i].bbmin);
			VectorScale (pseqdesc[i].bbmax, scale, pseqdesc[i].bbmax);
		}

		// maybe scale exeposition, pivots, attachments
	}

	/**
	 * {[XP[ݒ
	 */
	void scaleBones (float scale)
	{
		if (!m_pstudiohdr)
			return;

		mstudiobone_t *pbones = (mstudiobone_t *) ((byte *) m_pstudiohdr + m_pstudiohdr->boneindex);
		for (int i = 0; i < m_pstudiohdr->numbones; i++)
		{
			for (int j = 0; j < 3; j++)
			{
				pbones[i].value[j] *= scale;
				pbones[i].scale[j] *= scale;
			}
		}	
	}

	/**
	 * qbg{bNX`
	 */
	void drawBox (vec3_t *v)
	{
		SColor color = 0xFFFF0000;

		//ݒ
		SMaterial Material;
		Material.setFlag(EMF_LIGHTING, false);	//Cgݒ
		Material.GouraudShading = true;
		m_driver->setMaterial(Material); 

		//
		m_driver->draw3DLine(vector3df(v[6][0], v[6][2], v[6][1]), vector3df(v[7][0], v[7][2], v[7][1]), color);
		m_driver->draw3DLine(vector3df(v[0][0], v[0][2], v[0][1]), vector3df(v[1][0], v[1][2], v[1][1]), color);
		m_driver->draw3DLine(vector3df(v[4][0], v[4][2], v[4][1]), vector3df(v[5][0], v[5][2], v[5][1]), color);
		m_driver->draw3DLine(vector3df(v[2][0], v[2][2], v[2][1]), vector3df(v[3][0], v[3][2], v[3][1]), color);
		
		//c
		m_driver->draw3DLine(vector3df(v[6][0], v[6][2], v[6][1]), vector3df(v[0][0], v[0][2], v[0][1]), color);
		m_driver->draw3DLine(vector3df(v[0][0], v[0][2], v[0][1]), vector3df(v[2][0], v[2][2], v[2][1]), color);
		m_driver->draw3DLine(vector3df(v[4][0], v[4][2], v[4][1]), vector3df(v[6][0], v[6][2], v[6][1]), color);
		m_driver->draw3DLine(vector3df(v[2][0], v[2][2], v[2][1]), vector3df(v[4][0], v[4][2], v[4][1]), color);

		//c
		m_driver->draw3DLine(vector3df(v[1][0], v[1][2], v[1][1]), vector3df(v[7][0], v[7][2], v[7][1]), color);
		m_driver->draw3DLine(vector3df(v[7][0], v[7][2], v[7][1]), vector3df(v[5][0], v[5][2], v[5][1]), color);
		m_driver->draw3DLine(vector3df(v[3][0], v[3][2], v[3][1]), vector3df(v[1][0], v[1][2], v[1][1]), color);
		m_driver->draw3DLine(vector3df(v[5][0], v[5][2], v[5][1]), vector3df(v[3][0], v[3][2], v[3][1]), color);
	}

private:
	// entity settings
	vec3_t					m_origin;
	vec3_t					m_angles;	
	int						m_sequence;			// sequence index
	float					m_frame;			// frame
	int						m_bodynum;			// bodypart selection	
	int						m_skinnum;			// skin group selection
	byte					m_controller[4];	// bone controllers
	byte					m_blending[2];		// animation blending
	byte					m_mouth;			// mouth position
	bool					m_owntexmodel;		// do we have a modelT.mdl ?

	// internal data
	studiohdr_t				*m_pstudiohdr;
	mstudiomodel_t			*m_pmodel;

	studiohdr_t				*m_ptexturehdr;
	studiohdr_t				*m_panimhdr[32];

	vec4_t					m_adj;				// FIX: non persistant, make static

	/**
	 * {[ʒuvZ
	 */
	void CalcBoneAdj( void )
	{
		int					i, j;
		float				value;
		mstudiobonecontroller_t *pbonecontroller;
		
		pbonecontroller = (mstudiobonecontroller_t *)((byte *)m_pstudiohdr + m_pstudiohdr->bonecontrollerindex);

		for (j = 0; j < m_pstudiohdr->numbonecontrollers; j++)
		{
			i = pbonecontroller[j].index;
			if (i <= 3)
			{
				// check for 360% wrapping
				if (pbonecontroller[j].type & STUDIO_RLOOP)
				{
					value = m_controller[i] * (360.0/256.0) + pbonecontroller[j].start;
				}
				else 
				{
					value = m_controller[i] / 255.0;
					if (value < 0) value = 0;
					if (value > 1.0) value = 1.0;
					value = (1.0 - value) * pbonecontroller[j].start + value * pbonecontroller[j].end;
				}
				// Con_DPrintf( "%d %d %f : %f\n", m_controller[j], m_prevcontroller[j], value, dadt );
			}
			else
			{
				value = m_mouth / 64.0;
				if (value > 1.0) value = 1.0;
				value = (1.0 - value) * pbonecontroller[j].start + value * pbonecontroller[j].end;
				// Con_DPrintf("%d %f\n", mouthopen, value );
			}
			switch(pbonecontroller[j].type & STUDIO_TYPES)
			{
			case STUDIO_XR:
			case STUDIO_YR:
			case STUDIO_ZR:
				m_adj[j] = value * (Q_PI / 180.0);
				break;
			case STUDIO_X:
			case STUDIO_Y:
			case STUDIO_Z:
				m_adj[j] = value;
				break;
			}
		}
	}

	/**
	 * {[NH[^jIvZ
	 */
	void CalcBoneQuaternion( int frame, float s, mstudiobone_t *pbone, mstudioanim_t *panim, float *q )
	{
		int					j, k;
		vec4_t				q1, q2;
		vec3_t				angle1, angle2;
		mstudioanimvalue_t	*panimvalue;

		for (j = 0; j < 3; j++)
		{
			if (panim->offset[j+3] == 0)
			{
				angle2[j] = angle1[j] = pbone->value[j+3]; // default;
			}
			else
			{
				panimvalue = (mstudioanimvalue_t *)((byte *)panim + panim->offset[j+3]);
				k = frame;
				while (panimvalue->num.total <= k)
				{
					k -= panimvalue->num.total;
					panimvalue += panimvalue->num.valid + 1;
				}
				// Bah, missing blend!
				if (panimvalue->num.valid > k)
				{
					angle1[j] = panimvalue[k+1].value;

					if (panimvalue->num.valid > k + 1)
					{
						angle2[j] = panimvalue[k+2].value;
					}
					else
					{
						if (panimvalue->num.total > k + 1)
							angle2[j] = angle1[j];
						else
							angle2[j] = panimvalue[panimvalue->num.valid+2].value;
					}
				}
				else
				{
					angle1[j] = panimvalue[panimvalue->num.valid].value;
					if (panimvalue->num.total > k + 1)
					{
						angle2[j] = angle1[j];
					}
					else
					{
						angle2[j] = panimvalue[panimvalue->num.valid + 2].value;
					}
				}
				angle1[j] = pbone->value[j+3] + angle1[j] * pbone->scale[j+3];
				angle2[j] = pbone->value[j+3] + angle2[j] * pbone->scale[j+3];
			}

			if (pbone->bonecontroller[j+3] != -1)
			{
				angle1[j] += m_adj[pbone->bonecontroller[j+3]];
				angle2[j] += m_adj[pbone->bonecontroller[j+3]];
			}
		}

		if (!VectorCompare( angle1, angle2 ))
		{
			AngleQuaternion( angle1, q1 );
			AngleQuaternion( angle2, q2 );
			QuaternionSlerp( q1, q2, s, q );
		}
		else
		{
			AngleQuaternion( angle1, q );
		}
	}

	/**
	 * {[ʒuvZ
	 */
	void CalcBonePosition( int frame, float s, mstudiobone_t *pbone, mstudioanim_t *panim, float *pos )
	{
		int					j, k;
		mstudioanimvalue_t	*panimvalue;

		for (j = 0; j < 3; j++)
		{
			pos[j] = pbone->value[j]; // default;
			if (panim->offset[j] != 0)
			{
				panimvalue = (mstudioanimvalue_t *)((byte *)panim + panim->offset[j]);
				
				k = frame;
				// find span of values that includes the frame we want
				while (panimvalue->num.total <= k)
				{
					k -= panimvalue->num.total;
					panimvalue += panimvalue->num.valid + 1;
				}
				// if we're inside the span
				if (panimvalue->num.valid > k)
				{
					// and there's more data in the span
					if (panimvalue->num.valid > k + 1)
					{
						pos[j] += (panimvalue[k+1].value * (1.0 - s) + s * panimvalue[k+2].value) * pbone->scale[j];
					}
					else
					{
						pos[j] += panimvalue[k+1].value * pbone->scale[j];
					}
				}
				else
				{
					// are we at the end of the repeating values section and there's another section with data?
					if (panimvalue->num.total <= k + 1)
					{
						pos[j] += (panimvalue[panimvalue->num.valid].value * (1.0 - s) + s * panimvalue[panimvalue->num.valid + 2].value) * pbone->scale[j];
					}
					else
					{
						pos[j] += panimvalue[panimvalue->num.valid].value * pbone->scale[j];
					}
				}
			}
			if (pbone->bonecontroller[j] != -1)
			{
				pos[j] += m_adj[pbone->bonecontroller[j]];
			}
		}
	}

	/**
	 * ]vZ
	 */
	void CalcRotations ( vec3_t *pos, vec4_t *q, mstudioseqdesc_t *pseqdesc, mstudioanim_t *panim, float f )
	{
		int					i;
		int					frame;
		mstudiobone_t		*pbone;
		float				s;

		frame = (int)f;
		s = (f - frame);

		// add in programatic controllers
		CalcBoneAdj( );

		pbone		= (mstudiobone_t *)((byte *)m_pstudiohdr + m_pstudiohdr->boneindex);
		for (i = 0; i < m_pstudiohdr->numbones; i++, pbone++, panim++) 
		{
			CalcBoneQuaternion( frame, s, pbone, panim, q[i] );
			CalcBonePosition( frame, s, pbone, panim, pos[i] );
		}

		if (pseqdesc->motiontype & STUDIO_X)
			pos[pseqdesc->motionbone][0] = 0.0;
		if (pseqdesc->motiontype & STUDIO_Y)
			pos[pseqdesc->motionbone][1] = 0.0;
		if (pseqdesc->motiontype & STUDIO_Z)
			pos[pseqdesc->motionbone][2] = 0.0;
	}

	/**
	 * Aj[V擾
	 */
	mstudioanim_t *GetAnim( mstudioseqdesc_t *pseqdesc )
	{
		mstudioseqgroup_t	*pseqgroup;
		pseqgroup = (mstudioseqgroup_t *)((byte *)m_pstudiohdr + m_pstudiohdr->seqgroupindex) + pseqdesc->seqgroup;

		if (pseqdesc->seqgroup == 0)
		{
			return (mstudioanim_t *)((byte *)m_pstudiohdr + pseqgroup->data + pseqdesc->animindex);
		}

		return (mstudioanim_t *)((byte *)m_panimhdr[pseqdesc->seqgroup] + pseqdesc->animindex);
	}

	/**
	 * {[̋ʐ`
	 */
	void SlerpBones( vec4_t q1[], vec3_t pos1[], vec4_t q2[], vec3_t pos2[], float s )
	{
		int			i;
		vec4_t		q3;
		float		s1;

		if (s < 0) s = 0;
		else if (s > 1.0) s = 1.0;

		s1 = 1.0 - s;

		for (i = 0; i < m_pstudiohdr->numbones; i++)
		{
			QuaternionSlerp( q1[i], q2[i], s, q3 );
			q1[i][0] = q3[0];
			q1[i][1] = q3[1];
			q1[i][2] = q3[2];
			q1[i][3] = q3[3];
			pos1[i][0] = pos1[i][0] * s1 + pos2[i][0] * s;
			pos1[i][1] = pos1[i][1] * s1 + pos2[i][1] * s;
			pos1[i][2] = pos1[i][2] * s1 + pos2[i][2] * s;
		}
	}

	/**
	 * {[̐ݒ
	 */
	void SetUpBones ( void )
	{
		int					i;

		mstudiobone_t		*pbones;
		mstudioseqdesc_t	*pseqdesc;
		mstudioanim_t		*panim;

		static vec3_t		pos[MAXSTUDIOBONES];
		float				bonematrix[3][4];
		static vec4_t		q[MAXSTUDIOBONES];

		static vec3_t		pos2[MAXSTUDIOBONES];
		static vec4_t		q2[MAXSTUDIOBONES];
		static vec3_t		pos3[MAXSTUDIOBONES];
		static vec4_t		q3[MAXSTUDIOBONES];
		static vec3_t		pos4[MAXSTUDIOBONES];
		static vec4_t		q4[MAXSTUDIOBONES];


		if (m_sequence >=  m_pstudiohdr->numseq) {
			m_sequence = 0;
		}

		pseqdesc = (mstudioseqdesc_t *)((byte *)m_pstudiohdr + m_pstudiohdr->seqindex) + m_sequence;

		panim = GetAnim( pseqdesc );
		CalcRotations( pos, q, pseqdesc, panim, m_frame );

		if (pseqdesc->numblends > 1)
		{
			float				s;

			panim += m_pstudiohdr->numbones;
			CalcRotations( pos2, q2, pseqdesc, panim, m_frame );
			s = m_blending[0] / 255.0;

			SlerpBones( q, pos, q2, pos2, s );

			if (pseqdesc->numblends == 4)
			{
				panim += m_pstudiohdr->numbones;
				CalcRotations( pos3, q3, pseqdesc, panim, m_frame );

				panim += m_pstudiohdr->numbones;
				CalcRotations( pos4, q4, pseqdesc, panim, m_frame );

				s = m_blending[0] / 255.0;
				SlerpBones( q3, pos3, q4, pos4, s );

				s = m_blending[1] / 255.0;
				SlerpBones( q, pos, q3, pos3, s );
			}
		}

		pbones = (mstudiobone_t *)((byte *)m_pstudiohdr + m_pstudiohdr->boneindex);

		for (i = 0; i < m_pstudiohdr->numbones; i++) {
			QuaternionMatrix( q[i], bonematrix );

			bonematrix[0][3] = pos[i][0];
			bonematrix[1][3] = pos[i][1];
			bonematrix[2][3] = pos[i][2];

			if (pbones[i].parent == -1) {
				memcpy(m_bonetransform[i], bonematrix, sizeof(float) * 12);
			} 
			else {
				R_ConcatTransforms (m_bonetransform[pbones[i].parent], bonematrix, m_bonetransform[i]);
			}
		}
	}

	/**
	 * CeBOݒ
	 */
	void Lighting (float *lv, int bone, int flags, vec3_t normal)
	{
		float 	illum;
		float	lightcos;

		illum = m_ambientlight;

		if (flags & STUDIO_NF_FLATSHADE)
		{
			illum += m_shadelight * 0.8;
		} 
		else 
		{
			float r;
			lightcos = DotProduct (normal, m_blightvec[bone]); // -1 colinear, 1 opposite

			if (lightcos > 1.0)
				lightcos = 1;

			illum += m_shadelight;

			r = 1.5;//m_lambert;
			if (r <= 1.0) r = 1.0;

			lightcos = (lightcos + (r - 1.0)) / r; 		// do modified hemispherical lighting
			if (lightcos > 0.0) 
			{
				illum -= m_shadelight * lightcos; 
			}
			if (illum <= 0)
				illum = 0;
		}

		if (illum > 255) 
			illum = 255;
		*lv = illum / 255.0;	// Light from 0 to 1.0
	}

	/**
	 * Nݒ
	 */
	void Chrome (int *pchrome, int bone, vec3_t normal)
	{
		float n;

		if (m_chromeage[bone] != m_smodels_total)
		{
			vec3_t vright = { 50, 50, 0 };//{̓O[o

			// calculate vectors from the viewer to the bone. This roughly adjusts for position
			vec3_t chromeupvec;		// m_chrome t vector in world reference frame
			vec3_t chromerightvec;	// m_chrome s vector in world reference frame
			vec3_t tmp;				// vector pointing at bone in world reference frame
			VectorScale( m_origin, -1, tmp );
			tmp[0] += m_bonetransform[bone][0][3];
			tmp[1] += m_bonetransform[bone][1][3];
			tmp[2] += m_bonetransform[bone][2][3];
			VectorNormalize( tmp );
			CrossProduct( tmp, vright, chromeupvec );
			VectorNormalize( chromeupvec );
			CrossProduct( tmp, chromeupvec, chromerightvec );
			VectorNormalize( chromerightvec );

			VectorIRotate( chromeupvec, m_bonetransform[bone], m_chromeup[bone] );
			VectorIRotate( chromerightvec, m_bonetransform[bone], m_chromeright[bone] );

			m_chromeage[bone] = m_smodels_total;
		}

		// calc s coord
		n = DotProduct( normal, m_chromeright[bone] );
		pchrome[0] = (n + 1.0) * 32; // FIX: make this a float

		// calc t coord
		n = DotProduct( normal, m_chromeup[bone] );
		pchrome[1] = (n + 1.0) * 32; // FIX: make this a float
	}

	/**
	 * CeBOݒ
	 */
	void SetupLighting( void )
	{
		int i;
		m_ambientlight = 32;
		m_shadelight = 192;

		m_lightvec[0] = 0;
		m_lightvec[1] = 0;
		m_lightvec[2] = -1.0;

		m_lightcolor[0] = g_viewerSettings.lColor[0];
		m_lightcolor[1] = g_viewerSettings.lColor[1];
		m_lightcolor[2] = g_viewerSettings.lColor[2];

		// TODO: only do it for bones that actually have textures
		for (i = 0; i < m_pstudiohdr->numbones; i++)
		{
			VectorIRotate( m_lightvec, m_bonetransform[i], m_blightvec[i] );
		}
	}

	/**
	 * fݒ
	 */
	void SetupModel ( int bodypart )
	{
		int index;

		if (bodypart > m_pstudiohdr->numbodyparts)
		{
			// Con_DPrintf ("StudioModel::SetupModel: no such bodypart %d\n", bodypart);
			bodypart = 0;
		}

		mstudiobodyparts_t   *pbodypart = (mstudiobodyparts_t *)((byte *)m_pstudiohdr + m_pstudiohdr->bodypartindex) + bodypart;

		index = m_bodynum / pbodypart->base;
		index = index % pbodypart->nummodels;

		m_pmodel = (mstudiomodel_t *)((byte *)m_pstudiohdr + pbodypart->modelindex) + index;
	}

	/**
	 * f̕`{
	 */
	void DrawPoints( void )
	{
		int					i, j;
		mstudiomesh_t		*pmesh;
		byte				*pvertbone;
		byte				*pnormbone;
		vec3_t				*pstudioverts;
		vec3_t				*pstudionorms;
		mstudiotexture_t	*ptexture;
		float 				*av;
		float				*lv;
		float				lv_tmp;
		short				*pskinref;

		pvertbone = ((byte *)m_pstudiohdr + m_pmodel->vertinfoindex);
		pnormbone = ((byte *)m_pstudiohdr + m_pmodel->norminfoindex);
		ptexture = (mstudiotexture_t *)((byte *)m_ptexturehdr + m_ptexturehdr->textureindex);

		pmesh = (mstudiomesh_t *)((byte *)m_pstudiohdr + m_pmodel->meshindex);

		pstudioverts = (vec3_t *)((byte *)m_pstudiohdr + m_pmodel->vertindex);
		pstudionorms = (vec3_t *)((byte *)m_pstudiohdr + m_pmodel->normindex);

		pskinref = (short *)((byte *)m_ptexturehdr + m_ptexturehdr->skinindex);
		if (m_skinnum != 0 && m_skinnum < m_ptexturehdr->numskinfamilies)
			pskinref += (m_skinnum * m_ptexturehdr->numskinref);

		//_̕ύX
		for (i = 0; i < m_pmodel->numverts; i++)
		{
			vec3_t tmp;
			VectorScale (pstudioverts[i], 12, tmp);
			VectorTransform (pstudioverts[i], m_bonetransform[pvertbone[i]], m_pxformverts[i]);
		}

		//
		// clip and draw all triangles
		//
		//|S`
		lv = (float *)m_pvlightvalues;
		for (j = 0; j < m_pmodel->nummesh; j++) 
		{
			int flags;
			flags = ptexture[pskinref[pmesh[j].skinref]].flags;
			for (i = 0; i < pmesh[j].numnorms; i++, lv += 3, pstudionorms++, pnormbone++)
			{
				Lighting (&lv_tmp, *pnormbone, flags, (float *)pstudionorms);

				// FIX: move this check out of the inner loop
				if (flags & STUDIO_NF_CHROME)
					Chrome( m_chrome[(float (*)[3])lv - m_pvlightvalues], *pnormbone, (float *)pstudionorms );

				lv[0] = lv_tmp * m_lightcolor[0];
				lv[1] = lv_tmp * m_lightcolor[1];
				lv[2] = lv_tmp * m_lightcolor[2];
			}
		}

		for(j = 0; j < m_pmodel->nummesh; j++) 
		{
			float s, t;
			short		*ptricmds;

			pmesh = (mstudiomesh_t *)((byte *)m_pstudiohdr + m_pmodel->meshindex) + j;
			ptricmds = (short *)((byte *)m_pstudiohdr + pmesh->triindex);

			s = 1.0/(float)ptexture[pskinref[pmesh->skinref]].width;
			t = 1.0/(float)ptexture[pskinref[pmesh->skinref]].height;

			//eNX`ݒ
			ITexture* Texture;
			Texture = m_driver->getTexture(ptexture[pskinref[pmesh->skinref]].name);
			SMaterial Material, baseMaterial;
			Material.Lighting = false;
		//	Material.BackfaceCulling = false;//\̋ʕ@ĂȂ̂ŗʕ`=false
			Material.ZBuffer = true;

			//At@ݒ
			if(g_viewerSettings.transparency < 1.0f)
				//Material.MaterialType = EMT_TRANSPARENT_VERTEX_ALPHA;//_At@
				Material.MaterialType = EMT_NORMAL_MAP_TRANSPARENT_VERTEX_ALPHA;//_At@

			//`惂[hݒ
			switch(g_viewerSettings.renderMode)
			{
				case RM_WIREFRAME:
					Material.Wireframe = true;
					break;
				case RM_FLATSHADED:
					Material.GouraudShading = false;
					break;
				case RM_SMOOTHSHADED:
					Material.GouraudShading = true;
					break;
				case RM_TEXTURED:
					Material.TextureLayer[0].Texture = Texture;
					break;
			}

			m_driver->setMaterial(Material);

			bool bFan = false;
			S3DVertex ver[64];
			SColor color;
			u8 subCol, sub[3]; 
			s32 num = 0;
			//_ǂ邩

			//FANFԂɕ`悷̂ŁA_Lq邾őv
			u16 indexFan[] = {
				 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
				10,11,12,13,14,15,16,17,18,19,
				20,21,22,23,24,25,26,27,28,29,
				30,31,32,33,34,35,36,37,38,39,
				40,41,42,43,44,45,46,47,48,49,
				50,51,52,53,54,55,56,57,58,59,
				60,61,62,63,64,65,66,67,68,69,
			};

			//LISTF3̒_ŎOp``悷̂ŁA012,132ƋLq
			u16 index[] = {
				 0, 1, 2,  1, 3, 2,  2, 3, 4,  3, 5, 4,  4, 5, 6,  5, 7, 6,  6, 7, 8,  7, 9, 8,  8, 9,10,  9,11,10, 
				10,11,12, 11,13,12, 12,13,14, 13,15,14, 14,15,16, 15,17,16, 16,17,18, 17,19,18, 18,19,20, 19,21,20,
				20,21,22, 21,23,22, 22,23,24, 23,25,24, 24,25,26, 25,27,26, 26,27,28, 27,29,28, 28,29,30, 29,31,30, 
				30,31,32, 31,33,32, 32,33,34, 33,35,34, 34,35,36, 35,37,36, 36,37,38, 37,39,38, 38,39,40, 39,41,40,
				40,41,42, 41,43,42, 42,43,44, 43,45,44, 44,45,46, 45,47,46, 46,47,48, 47,49,48, 48,49,50, 49,51,50,
				50,51,52, 51,53,52, 52,53,54, 53,55,54, 54,55,56, 55,57,56, 56,57,58, 57,59,58, 58,59,60, 59,61,60,
				60,61,62, 61,63,62, 62,63,64, 63,65,64, 64,65,66, 65,67,66, 66,67,68, 67,69,68, 68,69,70, 69,71,70,
			};

			
			while (i = *(ptricmds++))
			{
				bFan = false;

				//̒lTRIANGLEFANŕ`悷
				if(i < 0)
				{
					i = -i;
					bFan = true;
				}

				num = 0;
				for( ; i > 0; i--, ptricmds += 4)
				{
					av = m_pxformverts[ptricmds[0]];	//_z
					lv = m_pvlightvalues[ptricmds[1]];	//Fz
					subCol = (g_viewerSettings.transparency * 255);
					sub[0] = lv[2]*255;
					sub[1] = lv[1]*255;
					sub[2] = lv[0]*255;
					color = subCol<<24 | sub[0]<<16 | sub[1]<<8 | sub[2];
					ver[num++] = S3DVertex(av[0], av[2], av[1],  0,0,0, color, ptricmds[2]*s, ptricmds[3]*t);
				}
				if(bFan)
					m_driver->drawIndexedTriangleFan(ver, num, indexFan, num-2);
				else
					m_driver->drawIndexedTriangleList(ver, num, index, num-2);
			}
			

			//ɖ߂
			m_driver->setMaterial(baseMaterial);
		}
	}
};

#endif // INCLUDED_STUDIOMODEL