/** --------------------------------------------------------------
 * IrrlichtΉ
 * author:         IrrLesson
 * date:           2008/04/27
 * email:          lesson@live.jp
 * web:            http://lesson.ifdef.jp/
 * --------------------------------------------------------------
 */

#ifdef WIN32
	#include <windows.h>
	#pragma comment(lib, "Irrlicht.lib")
#endif

#include <irrlicht.h>
#include <locale.h>
#include "mdlviewer.h"

using namespace irr;
using namespace gui;
using namespace video;

/**
 * [U[CxgID`
 */
enum USER_EVENT
{
	UID_NONE,

	//j[
	UID_EXIT,
	UID_OPEN,

	//^u1
	UID_COMBO_RANDER,
	UID_SCROLL_ALPHA,
	UID_CHECK_ATTACH,
	UID_CHECK_BONES,
	UID_CHECK_GROUND,
	UID_CHECK_MIRROR,
	UID_CHECK_BACK,
	UID_CHECK_HITBOX,
	UID_EDIT_SCALEX,
	UID_EDIT_SCALEY,
	UID_BUTTON_SCALEX,
	UID_BUTTON_SCALEY,

	//^u2
	UID_COMBO_ANIME,
	UID_SCROLL_SPEED,
	UID_BUTTON_STOP,
	UID_BUTTON_PREV,
	UID_BUTTON_NEXT,
	UID_EDIT_FRAME,

	//^u3
	UID_COMBO_STUDIO,
	UID_COMBO_UNKNOWN,
	UID_COMBO_SKIN,
	UID_COMBO_SUBMODEL,
	UID_SCROLL_UNKNOWN,

	//^u4
	UID_COMBO_BMP,
	UID_CHECK_CHROME,
	UID_SCROLL_CHROME,
	UID_BUTTON_EXPORT,
	UID_BUTTON_IMPORT,
	UID_BUTTON_SAVE,

	//^u5
	UID_EDIT_WIDTH,
	UID_EDIT_HEIGHT,
	UID_BUTTON_FULLSCREEN,

	//^u6
	UID_TEXT_BACK,
	UID_SCROLL_BACK_RED,
	UID_SCROLL_BACK_GREEN,
	UID_SCROLL_BACK_BLUE,
	UID_TEXT_LIGHT,
	UID_SCROLL_LIGHT_RED,
	UID_SCROLL_LIGHT_GREEN,
	UID_SCROLL_LIGHT_BLUE,
};

class CDialog
{
private:
	IVideoDriver*    m_driver;
	IGUIEnvironment* m_env;
	MDLViewer*       m_viewer;

	IGUIComboBox* boxAnim;

public:
	/**
	 * RXgN^
	 */
	CDialog(IVideoDriver* driver, IGUIEnvironment* env, MDLViewer* viewer)
		: m_driver(driver), m_env(env), m_viewer(viewer)
	{
		init();
	}

	/**
	 * fXgN^
	 */
	~CDialog()
	{
	}

	/**
	 * _CAOACe쐬
	 */
	void init()
	{
		//tHg蓖
		IGUISkin* skin = m_env->getSkin();
		IGUIFont* font = m_env->getFont("font.bmp");
		if(font)
			skin->setFont(font);

		//j[ݒ
		IGUIContextMenu* menu = m_env->addMenu();
		menu->addItem(L"File",    -1, true, true);
		menu->addItem(L"Options", -1, true, true);
		menu->addItem(L"View",    -1, true, true);
		menu->addItem(L"Help",    -1, true, true);

		//Tuj[ݒ
		gui::IGUIContextMenu* submenu;
		submenu = menu->getSubMenu(0);
		submenu->addItem(L"Load Model...", UID_OPEN);
		submenu->addSeparator();
		submenu->addItem(L"Exit", UID_EXIT);

		submenu = menu->getSubMenu(1);
		submenu->addItem(L"Background Color...", 100);
		submenu->addItem(L"Ground Color...", 100);
		submenu->addItem(L"Light Color...", 100);
		submenu->addSeparator();
		submenu->addItem(L"Center View...", 100);
		submenu->addSeparator();
		submenu->addItem(L"Make Screenshot...", 100);

		submenu = menu->getSubMenu(2);
		submenu->addItem(L"File Associations...", 100);

		submenu = menu->getSubMenu(3);
		submenu->addItem(L"Goto Homepage...", 100);
		submenu->addSeparator();
		submenu->addItem(L"About...", 100);

		//Rg[̃At@
		for(s32 i=0; i<EGDC_COUNT ; ++i)
		{
			SColor col = m_env->getSkin()->getColor((EGUI_DEFAULT_COLOR)i);
			col.setAlpha(255);
			m_env->getSkin()->setColor((EGUI_DEFAULT_COLOR)i, col);
		}
		m_env->getSkin()->setColor(EGDC_3D_SHADOW, 	m_env->getSkin()->getColor(EGDC_3D_FACE));
		//m_env->getSkin()->setColor(EGDC_3D_DARK_SHADOW, m_env->getSkin()->getColor(EGDC_3D_FACE));

		//ʉ̃^uRg[
		dimension2d<s32> size = m_driver->getScreenSize();
		IGUITabControl*   tab = m_env->addTabControl(
			core::rect<s32>(0,size.Height-150,size.Width,size.Height), 0, true, true);//ʃTCYŒ
		IGUITab* t1 = tab->addTab(L"Render");
		IGUITab* t2 = tab->addTab(L"Sequence");
		IGUITab* t3 = tab->addTab(L"Body");
		IGUITab* t4 = tab->addTab(L"Texture");
		IGUITab* t5 = tab->addTab(L"Fullscreen");
		IGUITab* t6 = tab->addTab(L"Color");

		//^u1y[W
		m_env->addCheckBox(false, rect<s32>( 10,60,150,60+20), t1, UID_CHECK_ATTACH, L"Attachments");
		m_env->addCheckBox(false, rect<s32>( 10,85,150,85+20), t1, UID_CHECK_BONES,  L"Bones");
		m_env->addCheckBox(false, rect<s32>(160,10,400,10+20), t1, UID_CHECK_GROUND, L"Ground");
		m_env->addCheckBox(false, rect<s32>(160,35,400,35+20), t1, UID_CHECK_MIRROR, L"Mirror Model On Ground");
		m_env->addCheckBox(false, rect<s32>(160,60,400,60+20), t1, UID_CHECK_BACK,   L"Background");
		m_env->addCheckBox(false, rect<s32>(160,85,400,85+20), t1, UID_CHECK_HITBOX, L"Hit Boxes");
		m_env->addEditBox(L"1.0", rect<s32>(410,10,460,10+20), true, t1, UID_EDIT_SCALEX);
		m_env->addEditBox(L"1.0", rect<s32>(410,35,460,35+20), true, t1, UID_EDIT_SCALEY);
		m_env->addButton(rect<s32>(470,10,560,10+20), t1, UID_BUTTON_SCALEX, L"Scale");
		m_env->addButton(rect<s32>(470,35,560,35+20), t1, UID_BUTTON_SCALEY, L"Scale");

		IGUIComboBox* box = m_env->addComboBox(rect<s32>(10,10,150,10+20), t1, UID_COMBO_RANDER);
		box->addItem(L"Wireframe");
		box->addItem(L"Flatshaded");
		box->addItem(L"Smoothshaded");
		box->addItem(L"Textured");
		box->setSelected(3);//TexturedI

		IGUIScrollBar* scrollbar = m_env->addScrollBar(true, rect<s32>(10,35,150,35+20), t1, UID_SCROLL_ALPHA);
		scrollbar->setMax(255);
		scrollbar->setPos(255);

		//^u2y[W
		boxAnim = m_env->addComboBox(rect<s32>(10,10,240,10+20), t2, UID_COMBO_ANIME);
		scrollbar = m_env->addScrollBar(true, rect<s32>(10,35,240,35+20), t2, UID_SCROLL_SPEED);
		scrollbar->setMax(255);
		scrollbar->setPos(40);

		m_env->addButton(rect<s32>( 10,60, 80,60+20), t2, UID_BUTTON_SCALEX, L"Stop");//"Play"
		m_env->addButton(rect<s32>( 90,60,130,60+20), t2, UID_BUTTON_SCALEX, L"<<");
		m_env->addButton(rect<s32>(200,60,240,60+20), t2, UID_BUTTON_SCALEX, L">>");
		m_env->addEditBox(L"", rect<s32>(140,60,190,60+20), true, t2, UID_EDIT_FRAME);

		//^u3y[W
		box = m_env->addComboBox(rect<s32>(10,10,100,10+20), t3, UID_COMBO_STUDIO);
		box->addItem(L"studio");
		box = m_env->addComboBox(rect<s32>(10,35,100,35+20), t3, UID_COMBO_UNKNOWN);
		box->addItem(L"");
		box->setEnabled(false);
		box = m_env->addComboBox(rect<s32>(10,60,100,60+20), t3, UID_COMBO_SKIN);
		box->addItem(L"Skin 1");

		box = m_env->addComboBox(rect<s32>(110,10,200,10+20), t3, UID_COMBO_SUBMODEL);
		box->addItem(L"Submodel 1");
		scrollbar = m_env->addScrollBar(true, rect<s32>(110,35,200,35+20), t3, UID_SCROLL_UNKNOWN);
		scrollbar->setMax(255);
		scrollbar->setPos(0);

		//^u4y[W
		box = m_env->addComboBox(rect<s32>(10,10,240,10+20), t4, UID_COMBO_BMP);
		m_env->addCheckBox(false,rect<s32>(10,35,240,35+20), t4, UID_CHECK_CHROME,  L"Chrome Effect");
		scrollbar = m_env->addScrollBar(true, rect<s32>(10,60,240,60+20), t4, UID_SCROLL_CHROME);
		scrollbar->setMax(255);
		scrollbar->setPos(0);

		m_env->addButton(rect<s32>(250,10,350,10+20), t4, UID_BUTTON_EXPORT, L"Export");//"Play"
		m_env->addButton(rect<s32>(250,35,350,35+20), t4, UID_BUTTON_IMPORT, L"Import");
		m_env->addButton(rect<s32>(250,60,350,60+20), t4, UID_BUTTON_SAVE,   L"Save Model");

		//^u5y[W
		m_env->addEditBox(L"800", rect<s32>( 10,10, 90,10+20), true, t5, UID_EDIT_WIDTH);
		m_env->addEditBox(L"600", rect<s32>(110,10,190,10+20), true, t5, UID_EDIT_HEIGHT);
		m_env->addButton(rect<s32>(200,10,300,10+20), t5, UID_BUTTON_FULLSCREEN, L"Fullscreen");

		//^u6y[W
		m_env->addStaticText(L"Back Color",  rect<s32>(10,10,240,10+20), false, true, t6, UID_TEXT_BACK);
		scrollbar = m_env->addScrollBar(true, rect<s32>(10,35,240,35+20), t6, UID_SCROLL_BACK_RED);
		scrollbar->setMax(255);
		scrollbar->setPos(0);
		scrollbar = m_env->addScrollBar(true, rect<s32>(10,60,240,60+20), t6, UID_SCROLL_BACK_GREEN);
		scrollbar->setMax(255);
		scrollbar->setPos(0);
		scrollbar = m_env->addScrollBar(true, rect<s32>(10,85,240,85+20), t6, UID_SCROLL_BACK_BLUE);
		scrollbar->setMax(255);
		scrollbar->setPos(0);

		m_env->addStaticText(L"Light Color",  rect<s32>(250,10,480,10+20), false, true, t6, UID_TEXT_LIGHT);
		scrollbar = m_env->addScrollBar(true, rect<s32>(250,35,480,35+20), t6, UID_SCROLL_LIGHT_RED);
		scrollbar->setMax(255);
		scrollbar->setPos(255);
		scrollbar = m_env->addScrollBar(true, rect<s32>(250,60,480,60+20), t6, UID_SCROLL_LIGHT_GREEN);
		scrollbar->setMax(255);
		scrollbar->setPos(255);
		scrollbar = m_env->addScrollBar(true, rect<s32>(250,85,480,85+20), t6, UID_SCROLL_LIGHT_BLUE);
		scrollbar->setMax(255);
		scrollbar->setPos(255);
	}


	/**
	 * _CAOACeXV
	 */
	void update()
	{
		studiohdr_t *hdr = m_viewer->getStudioHeader();
		mstudioseqdesc_t *pseqdesc;

		boxAnim->clear();

		s32 i,j;
		unsigned short wstr[32];

		if(hdr == NULL)
			return;

		for(i=0;i<hdr->numseq;i++)
		{
			pseqdesc = (mstudioseqdesc_t *)((byte *)hdr + hdr->seqindex) + i;
			for(j=0;j<32;j++)
				wstr[j] = pseqdesc->label[j];
			boxAnim->addItem(wstr);
		}
	}
};


/**
 * CxgV[o
 */
class MyEventReceiver : public IEventReceiver
{
	#define ZOOM_VALUE 5
	#define ROT_VALUE 0.05f

	IGUIEnvironment* m_env;
	MDLViewer*       m_viewer;
	IGUIWindow*      m_window;
	CDialog*         m_dialog;
	SColor           m_backColor;
	bool             m_move;
	s32              m_zoom;
	s32              m_oldX, m_oldY, m_oldZ;
	f64              m_rotX, m_rotZ;
	bool             m_lPush, m_rPush;

public:
	bool m_exit;

	/**
	 * RXgN^
	 * @param env    GUI
	 * @param viewer r[A
	 * @param dialog r[Ap_CAO
	 */
	MyEventReceiver(IGUIEnvironment* env, MDLViewer* viewer, CDialog* dialog)
		: m_env(env), m_viewer(viewer), m_dialog(dialog)
	{
		m_exit  = false;
		m_zoom  = 30;
		m_move  = false;
		m_backColor = 0xFF000000;

		m_rPush = false;
		m_lPush = false;

		m_rotX = m_rotZ = 0;
	}

	/** 
	 * fXgN^
	 */
	~MyEventReceiver()
	{
	}

	/**
	 * wiF擾
	 * @return wiF
	 */
	SColor getBackColor()
	{
		return m_backColor;
	}

	/**
	 * gkl擾
	 * @return gkl
	 */
	s32 getZoom()
	{
		return m_zoom;
	}

	/**
	 * ]l擾
	 * @return ]l
	 */
	vector3df getRotation()
	{
		return vector3df(0,m_rotX,m_rotZ);
	}

	/**
	 * Cxg
	 * @param event Cxg
	 * @return 0:ȂA1:
	 */
	virtual bool OnEvent(const SEvent& event)
	{
		//}EXCxg
		if(event.EventType == EET_MOUSE_INPUT_EVENT)
		{
			switch(event.MouseInput.Event)
			{
				//{^
				case EMIE_LMOUSE_PRESSED_DOWN:
					m_lPush = true;
					break;
				case EMIE_LMOUSE_LEFT_UP:
					m_lPush = false;
					break;

				//E{^
				case EMIE_RMOUSE_PRESSED_DOWN:
					m_rPush = true;
					break;
				case EMIE_RMOUSE_LEFT_UP:
					m_rPush = false;
					break;

				//zC[
				case EMIE_MOUSE_WHEEL:
				//	if(event.MouseInput.Wheel == 1){]
				//	else if(event.MouseInput.Wheel == -1){}
					break;
			}

			//{^Ă
			if(m_lPush)
			{
				if(event.MouseInput.X < m_oldX)
					m_rotX += ROT_VALUE;
				else if(event.MouseInput.X > m_oldX)
					m_rotX -= ROT_VALUE;

				if(event.MouseInput.Y < m_oldZ)
					m_rotZ += ROT_VALUE;
				else if(event.MouseInput.Y > m_oldZ)
					m_rotZ -= ROT_VALUE;

				m_oldX = event.MouseInput.X;
				m_oldZ = event.MouseInput.Y;

			}
			//E{^Ă
			if(m_rPush)
			{
				if(event.MouseInput.Y < m_oldY)
					m_zoom -= ZOOM_VALUE;
				else if(event.MouseInput.Y > m_oldY)
					m_zoom += ZOOM_VALUE;

				m_oldY = event.MouseInput.Y;
			}
		}

		//GUĨCxg
		if(event.EventType == EET_GUI_EVENT)
		{
			s32 id = event.GUIEvent.Caller->getID();

			switch(event.GUIEvent.EventType)
			{
				//j[
				case EGET_MENU_ITEM_SELECTED:
					{
						IGUIContextMenu* menu = (IGUIContextMenu*)event.GUIEvent.Caller;
						id = menu->getItemCommandId(menu->getSelectedItem());

						switch(id)
						{
							case UID_EXIT:
								m_exit = true;
								break;
							case UID_OPEN:
								m_env->addFileOpenDialog(L"Please select a Model file to open");
								break;
						}
					}
					break;

				//t@CI[v
				case EGET_FILE_SELECTED:
					{
						IGUIFileOpenDialog* dialog = 
							(IGUIFileOpenDialog*)event.GUIEvent.Caller;
						m_viewer->loadModel(stringc(dialog->getFileName()).c_str());

						//_CAOXV
						m_dialog->update();
					}
					break;

				//`FbN{bNX
				case EGET_CHECKBOX_CHANGED:
					{
						bool bCheck = ((IGUICheckBox*)event.GUIEvent.Caller)->isChecked();

						switch(id)
						{
							case UID_CHECK_ATTACH:
								g_viewerSettings.showAttachments = bCheck;
								break;
							case UID_CHECK_BONES:
								g_viewerSettings.showBones = bCheck;
								break;
							case UID_CHECK_GROUND:
								g_viewerSettings.showGround = bCheck;
								break;
						//	case UID_CHECK_MIRROR:
						//		g_viewerSettings.showAttachments = bCheck;
						//		break;
							case UID_CHECK_BACK:
								g_viewerSettings.showBackground = bCheck;
								break;
							case UID_CHECK_HITBOX:
								g_viewerSettings.showHitBoxes = bCheck;
								break;
						}
					}
					break;

				//R{{bNX
				case EGET_COMBO_BOX_CHANGED:
					if(id == UID_COMBO_RANDER)
					{
						s32 value = ((IGUIComboBox*)event.GUIEvent.Caller)->getSelected();
						g_viewerSettings.renderMode = value;
					}
					else if(id == UID_COMBO_ANIME)
					{
						s32 value = ((IGUIComboBox*)event.GUIEvent.Caller)->getSelected();
						m_viewer->setSequence(value);
					}
					break;

				//XN[o[
				case EGET_SCROLL_BAR_CHANGED:
					s32 pos;
					switch(id)
					{
						//xύX
						case UID_SCROLL_ALPHA:
							pos = ((IGUIScrollBar*)event.GUIEvent.Caller)->getPos();
							g_viewerSettings.transparency = (f32)pos/255;
							break;

						//CgFύX
						case UID_SCROLL_LIGHT_BLUE:
							pos = ((IGUIScrollBar*)event.GUIEvent.Caller)->getPos();
							g_viewerSettings.lColor[0] = (f32)pos/255;//B
							break;
						case UID_SCROLL_LIGHT_GREEN:
							pos = ((IGUIScrollBar*)event.GUIEvent.Caller)->getPos();
							g_viewerSettings.lColor[1] = (f32)pos/255;//G
							break;
						case UID_SCROLL_LIGHT_RED:
							pos = ((IGUIScrollBar*)event.GUIEvent.Caller)->getPos();
							g_viewerSettings.lColor[2] = (f32)pos/255;//R
							break;

						//wiFύX
						case UID_SCROLL_BACK_BLUE:
							pos = ((IGUIScrollBar*)event.GUIEvent.Caller)->getPos();
							m_backColor.setBlue(pos);//B
							break;
						case UID_SCROLL_BACK_GREEN:
							pos = ((IGUIScrollBar*)event.GUIEvent.Caller)->getPos();
							m_backColor.setGreen(pos);//G
							break;
						case UID_SCROLL_BACK_RED:
							pos = ((IGUIScrollBar*)event.GUIEvent.Caller)->getPos();
							m_backColor.setRed(pos);//R
							break;
					}
					break;

				//{^
				case EGET_BUTTON_CLICKED:
					break;
			}
		}
		return false;
	}
};



/**
 * C
 */
#ifdef WIN32
	int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR strCmdLine, int nShowCmd)
#else
	int main()
#endif
{
	IrrlichtDevice*  device = createDevice(EDT_OPENGL,dimension2d<s32>(600,480),32,false,false,false);
	IVideoDriver*    driver = device->getVideoDriver();
	ISceneManager*   smgr   = device->getSceneManager();
	IGUIEnvironment* env    = device->getGUIEnvironment();

	device->setWindowCaption(L"Irrlicht");//EChE^Cgݒ

	//Jݒ
	ICameraSceneNode* cam = smgr->addCameraSceneNode(0, vector3df(30,-5,0), vector3df(0,-5,0));
	
	//r[A
	MDLViewer* viewer = new MDLViewer(device);

	//r[Ap_CAO쐬
	CDialog* dialog = new CDialog(driver, env, viewer);
	MyEventReceiver* recv   = new MyEventReceiver(env, viewer, dialog);
	device->setEventReceiver(recv);

	//FOɃt@C̓ǂݍ݂ƃAj[V̍Đ
	//viewer->loadModel("v_elite.mdl");
	//viewer->setSequence(3);

	while(device->run() && recv->m_exit == false)
	{
		driver->beginScene(true,true,recv->getBackColor());

		smgr->setActiveCamera(cam);
		cam->setPosition(vector3df(recv->getZoom(),-5,0));

		matrix4 rotate;
		rotate.setRotationRadians(recv->getRotation());
		driver->setTransform(ETS_WORLD, rotate);

		//r[A`
		viewer->drawAll();

		//V[̕`
		smgr->drawAll();

		//GUI`
		env->drawAll();

		driver->endScene();
	}

	driver->drop();

	return 0;
}