/** --------------------------------------------------------------
 * IrrlichtΉ
 * author:         IrrLesson
 * date:           2008/04/27
 * email:          lesson@live.jp
 * web:            http://lesson.ifdef.jp/
 * --------------------------------------------------------------
 */

/***
*
*	Copyright (c) 1998, Valve LLC. All rights reserved.
*	
*	This product contains software technology licensed from Id 
*	Software, Inc. ("Id Technology").  Id Technology (c) 1996 Id Software, Inc. 
*	All Rights Reserved.
*
****/

#ifndef __MATHLIB__
#define __MATHLIB__

#include <math.h>

#pragma warning( disable : 4244 )//double to float

#ifndef byte
	typedef unsigned char byte;
#endif

#ifdef DOUBLEVEC_T
	typedef double vec_t;
#else
	typedef float vec_t;
#endif

typedef vec_t vec3_t[3];	// x,y,z
typedef vec_t vec4_t[4];	// x,y,z,w

#define	SIDE_FRONT		0
#define	SIDE_ON			2
#define	SIDE_BACK		1
#define	SIDE_CROSS		-2

#define	Q_PI	3.14159265358979323846

//extern vec3_t vec3_origin;

// Use this definition globally
#define	ON_EPSILON		0.01
#define	EQUAL_EPSILON	0.001

#define DotProduct(x,y)       ((x)[0]*(y)[0]+(x)[1]*(y)[1]+(x)[2]*(y)[2])
#define VectorFill(a,b)       { (a)[0]=(b); (a)[1]=(b); (a)[2]=(b);}
#define VectorAvg(a)          ( ( (a)[0] + (a)[1] + (a)[2] ) / 3 )
#define VectorSubtract(a,b,c) {(c)[0]=(a)[0]-(b)[0];(c)[1]=(a)[1]-(b)[1];(c)[2]=(a)[2]-(b)[2];}
#define VectorAdd(a,b,c)      {(c)[0]=(a)[0]+(b)[0];(c)[1]=(a)[1]+(b)[1];(c)[2]=(a)[2]+(b)[2];}
#define VectorCopy(a,b)       {(b)[0]=(a)[0];(b)[1]=(a)[1];(b)[2]=(a)[2];}
#define VectorScale(a,b,c)    {(c)[0]=(b)*(a)[0];(c)[1]=(b)*(a)[1];(c)[2]=(b)*(a)[2];}

//#ifdef _MAINFUNC_
	/**
	 * xN^[̔r
	 */
	int VectorCompare (vec3_t v1, vec3_t v2)
	{
		int		i;
		
		for (i=0 ; i<3 ; i++)
			if (fabs(v1[i]-v2[i]) > EQUAL_EPSILON)
				return false;
				
		return true;
	}

	/**
	 * xN^[̐K
	 */
	vec_t VectorNormalize (vec3_t v)
	{
		int		i;
		double	length;

		if ( fabs(v[1] - 0.000215956) < 0.0001)
			i=1;

		length = 0;
		for (i=0 ; i< 3 ; i++)
			length += v[i]*v[i];
		length = sqrt (length);
		if (length == 0)
			return 0;
			
		for (i=0 ; i< 3 ; i++)
			v[i] /= length;	

		return length;
	}

	// rotate by the inverse of the matrix
	/**
	 * xN^[̉]
	 */
	void VectorIRotate (const vec3_t in1, const float in2[3][4], vec3_t out)
	{
		out[0] = in1[0]*in2[0][0] + in1[1]*in2[1][0] + in1[2]*in2[2][0];
		out[1] = in1[0]*in2[0][1] + in1[1]*in2[1][1] + in1[2]*in2[2][1];
		out[2] = in1[0]*in2[0][2] + in1[1]*in2[1][2] + in1[2]*in2[2][2];
	}

	/**
	 * xN^[̕ό`
	 */
	void VectorTransform (const vec3_t in1, const float in2[3][4], vec3_t out)
	{
		out[0] = DotProduct(in1, in2[0]) + in2[0][3];
		out[1] = DotProduct(in1, in2[1]) +	in2[1][3];
		out[2] = DotProduct(in1, in2[2]) +	in2[2][3];
	}

	/**
	 * 
	 */
	void R_ConcatTransforms (const float in1[3][4], const float in2[3][4], float out[3][4])
	{
		out[0][0] = in1[0][0] * in2[0][0] + in1[0][1] * in2[1][0] +
					in1[0][2] * in2[2][0];
		out[0][1] = in1[0][0] * in2[0][1] + in1[0][1] * in2[1][1] +
					in1[0][2] * in2[2][1];
		out[0][2] = in1[0][0] * in2[0][2] + in1[0][1] * in2[1][2] +
					in1[0][2] * in2[2][2];
		out[0][3] = in1[0][0] * in2[0][3] + in1[0][1] * in2[1][3] +
					in1[0][2] * in2[2][3] + in1[0][3];
		out[1][0] = in1[1][0] * in2[0][0] + in1[1][1] * in2[1][0] +
					in1[1][2] * in2[2][0];
		out[1][1] = in1[1][0] * in2[0][1] + in1[1][1] * in2[1][1] +
					in1[1][2] * in2[2][1];
		out[1][2] = in1[1][0] * in2[0][2] + in1[1][1] * in2[1][2] +
					in1[1][2] * in2[2][2];
		out[1][3] = in1[1][0] * in2[0][3] + in1[1][1] * in2[1][3] +
					in1[1][2] * in2[2][3] + in1[1][3];
		out[2][0] = in1[2][0] * in2[0][0] + in1[2][1] * in2[1][0] +
					in1[2][2] * in2[2][0];
		out[2][1] = in1[2][0] * in2[0][1] + in1[2][1] * in2[1][1] +
					in1[2][2] * in2[2][1];
		out[2][2] = in1[2][0] * in2[0][2] + in1[2][1] * in2[1][2] +
					in1[2][2] * in2[2][2];
		out[2][3] = in1[2][0] * in2[0][3] + in1[2][1] * in2[1][3] +
					in1[2][2] * in2[2][3] + in1[2][3];
	}

	/**
	 * 
	 */
	void CrossProduct (vec3_t v1, vec3_t v2, vec3_t cross)
	{
		cross[0] = v1[1]*v2[2] - v1[2]*v2[1];
		cross[1] = v1[2]*v2[0] - v1[0]*v2[2];
		cross[2] = v1[0]*v2[1] - v1[1]*v2[0];
	}

	/**
	 * px̃NH[^jI
	 */
	void AngleQuaternion( const vec3_t angles, vec4_t quaternion )
	{
		float		angle;
		float		sr, sp, sy, cr, cp, cy;

		// FIXME: rescale the inputs to 1/2 angle
		angle = angles[2] * 0.5;
		sy = sin(angle);
		cy = cos(angle);
		angle = angles[1] * 0.5;
		sp = sin(angle);
		cp = cos(angle);
		angle = angles[0] * 0.5;
		sr = sin(angle);
		cr = cos(angle);

		quaternion[0] = sr*cp*cy-cr*sp*sy; // X
		quaternion[1] = cr*sp*cy+sr*cp*sy; // Y
		quaternion[2] = cr*cp*sy-sr*sp*cy; // Z
		quaternion[3] = cr*cp*cy+sr*sp*sy; // W
	}

	/**
	 * NH[^jI}gNX
	 */
	void QuaternionMatrix( const vec4_t quaternion, float (*matrix)[4] )
	{
		matrix[0][0] = 1.0 - 2.0 * quaternion[1] * quaternion[1] - 2.0 * quaternion[2] * quaternion[2];
		matrix[1][0] = 2.0 * quaternion[0] * quaternion[1] + 2.0 * quaternion[3] * quaternion[2];
		matrix[2][0] = 2.0 * quaternion[0] * quaternion[2] - 2.0 * quaternion[3] * quaternion[1];

		matrix[0][1] = 2.0 * quaternion[0] * quaternion[1] - 2.0 * quaternion[3] * quaternion[2];
		matrix[1][1] = 1.0 - 2.0 * quaternion[0] * quaternion[0] - 2.0 * quaternion[2] * quaternion[2];
		matrix[2][1] = 2.0 * quaternion[1] * quaternion[2] + 2.0 * quaternion[3] * quaternion[0];

		matrix[0][2] = 2.0 * quaternion[0] * quaternion[2] + 2.0 * quaternion[3] * quaternion[1];
		matrix[1][2] = 2.0 * quaternion[1] * quaternion[2] - 2.0 * quaternion[3] * quaternion[0];
		matrix[2][2] = 1.0 - 2.0 * quaternion[0] * quaternion[0] - 2.0 * quaternion[1] * quaternion[1];
	}

	/**
	 * NH[^jIʐ`ԁiSpherical Linear intERPolationj
	 */
	void QuaternionSlerp( const vec4_t p, vec4_t q, float t, vec4_t qt )
	{
		int i;
		float omega, cosom, sinom, sclp, sclq;

		// decide if one of the quaternions is backwards
		float a = 0;
		float b = 0;
		for (i = 0; i < 4; i++) {
			a += (p[i]-q[i])*(p[i]-q[i]);
			b += (p[i]+q[i])*(p[i]+q[i]);
		}
		if (a > b) {
			for (i = 0; i < 4; i++) {
				q[i] = -q[i];
			}
		}

		cosom = p[0]*q[0] + p[1]*q[1] + p[2]*q[2] + p[3]*q[3];

		if ((1.0 + cosom) > 0.00000001) {
			if ((1.0 - cosom) > 0.00000001) {
				omega = acos( cosom );
				sinom = sin( omega );
				sclp = sin( (1.0 - t)*omega) / sinom;
				sclq = sin( t*omega ) / sinom;
			}
			else {
				sclp = 1.0 - t;
				sclq = t;
			}
			for (i = 0; i < 4; i++) {
				qt[i] = sclp * p[i] + sclq * q[i];
			}
		}
		else {
			qt[0] = -p[1];
			qt[1] = p[0];
			qt[2] = -p[3];
			qt[3] = p[2];
			sclp = sin( (1.0 - t) * 0.5 * Q_PI);
			sclq = sin( t * 0.5 * Q_PI);
			for (i = 0; i < 3; i++) {
				qt[i] = sclp * p[i] + sclq * qt[i];
			}
		}
	}
/*
#else
	int VectorCompare (vec3_t v1, vec3_t v2);
	vec_t VectorNormalize (vec3_t v);
	void VectorIRotate (const vec3_t in1, const float in2[3][4], vec3_t out);
	void VectorTransform (const vec3_t in1, const float in2[3][4], vec3_t out);
	void R_ConcatTransforms (const float in1[3][4], const float in2[3][4], float out[3][4]);
	void CrossProduct (vec3_t v1, vec3_t v2, vec3_t cross);
	void AngleQuaternion( const vec3_t angles, vec4_t quaternion );
	void QuaternionMatrix( const vec4_t quaternion, float (*matrix)[4] );
	void QuaternionSlerp( const vec4_t p, vec4_t q, float t, vec4_t qt );
#endif
*/

#endif
