/** --------------------------------------------------------------
 * IrrlichtΉ
 * author:         IrrLesson
 * date:           2008/04/27
 * email:          lesson@live.jp
 * web:            http://lesson.ifdef.jp/
 * --------------------------------------------------------------
 */

//
//                 Half-Life Model Viewer (c) 1999 by Mete Ciragan
//
// file:           mdlviewer.h
// last modified:  Apr 28 1999, Mete Ciragan
// copyright:      The programs and associated files contained in this
//                 distribution were developed by Mete Ciragan. The programs
//                 are not in the public domain, but they are freely
//                 distributable without licensing fees. These programs are
//                 provided without guarantee or warrantee expressed or
//                 implied.
//
// version:        1.2
//
// email:          mete@swissquake.ch
// web:            http://www.swissquake.ch/chumbalum-soft/
//

#ifndef INCLUDED_MDLVIEWER
#define INCLUDED_MDLVIEWER

#include "ControlPanel.h"

#define IDC_FILE_LOADMODEL			1001
#define IDC_FILE_LOADBACKGROUNDTEX	1002
#define IDC_FILE_LOADGROUNDTEX		1003
#define IDC_FILE_UNLOADGROUNDTEX	1004
#define IDC_FILE_OPENPAKFILE		1005
#define IDC_FILE_OPENPAKFILE2		1006
#define IDC_FILE_CLOSEPAKFILE		1007
#define IDC_FILE_RECENTMODELS1		1008
#define IDC_FILE_RECENTMODELS2		1009
#define IDC_FILE_RECENTMODELS3		1010
#define IDC_FILE_RECENTMODELS4		1011
#define IDC_FILE_RECENTPAKFILES1	1012
#define IDC_FILE_RECENTPAKFILES2	1013
#define IDC_FILE_RECENTPAKFILES3	1014
#define IDC_FILE_RECENTPAKFILES4	1015
#define IDC_FILE_EXIT				1016

#define IDC_OPTIONS_COLORBACKGROUND	1101
#define IDC_OPTIONS_COLORGROUND		1102
#define IDC_OPTIONS_COLORLIGHT		1103
#define IDC_OPTIONS_CENTERVIEW		1104
#define IDC_OPTIONS_MAKESCREENSHOT	1105
#define IDC_OPTIONS_FILTERTEXTURES	1106
#define IDC_OPTIONS_DUMP			1107

#define IDC_VIEW_FILEASSOCIATIONS	1201

#define IDC_HELP_GOTOHOMEPAGE		1301
#define IDC_HELP_ABOUT				1302


class MDLViewer : public ControlPanel
{
	IrrlichtDevice* m_device;
	IVideoDriver* m_driver;
	int d_textureNames[2];

public:
	/**
	 * RXgN^
	 */
	MDLViewer(IrrlichtDevice* device)
		: m_device(device), m_driver(device->getVideoDriver()), ControlPanel()
	{
		//ړnew
		m_studioModel = new StudioModel(m_driver);
	}

	/**
	 * fXgN^
	 */
	~MDLViewer()
	{
		loadTexture (0, 0);
		loadTexture (0, 1);
	}

	/**
	 * ҂ԏ
	 */
	int idle ()
	{
		m_studioModel->SetBlending (0, 0.0);
		m_studioModel->SetBlending (1, 0.0);

		static float prev;
		float curr = (float) m_device->getTimer()->getRealTime() / 1000.0f;//irrlicht

		//if (!g_bStopPlaying)
			m_studioModel->AdvanceFrame ((curr - prev) * g_viewerSettings.speedScale);

		prev = curr;

		return 1;
	}

	/**
	 * `
	 */
	void drawAll ()
	{
		idle ();

		if (g_viewerSettings.showTexture)
		{
			studiohdr_t *hdr = m_studioModel->getTextureHeader ();
			if (hdr)
			{
				mstudiotexture_t *ptextures = (mstudiotexture_t *) ((byte *) hdr + hdr->textureindex);
			//	float w = (float) ptextures[g_viewerSettings.texture].width  * g_viewerSettings.textureScale;
			//	float h = (float) ptextures[g_viewerSettings.texture].height * g_viewerSettings.textureScale;
			//	float x = ((float) 800 - w) / 2;
			//	float y = ((float) 600 - h) / 2;
			}
			return;
		}

//		g_vright[0] = g_vright[1] = g_viewerSettings.trans[2];

		//~[
		if (g_viewerSettings.mirror)
		{
			m_studioModel->DrawModel ();
		}

		//f`
		m_studioModel->DrawModel ();

		//nʕ`
		if (g_viewerSettings.showGround)
		{
			drawFloor();
		}
	}

	/**
	 * X^WIwb_擾
	 */
	studiohdr_t *getStudioHeader()
	{
		return m_studioModel->getStudioHeader();
	}

	/**
	 * eNX`ǂݍ
	 */
	int loadTexture (const char *filename, int name)
	{
		if (!filename || !strlen (filename))
		{
			if (d_textureNames[name])
			{
	//			glDeleteTextures (1, (const GLuint *) &d_textureNames[name]);
				d_textureNames[name] = 0;

				if (name == 0)
					strcpy (g_viewerSettings.backgroundTexFile, "");
				else
					strcpy (g_viewerSettings.groundTexFile, "");
			}
			return 0;
		}

		return 0;
	}

	/**
	 * `
	 */
	void drawFloor ()
	{
		//}eAݒ
		SMaterial Material, baseMaterial;
		Material.Lighting = false;
		Material.BackfaceCulling = false;
		Material.MaterialType = EMT_TRANSPARENT_ADD_COLOR;//EMT_TRANSPARENT_VERTEX_ALPHA;//_ƃoO
		m_driver->setMaterial(Material);

		unsigned short index[] = {0,1,2, 0,2,3};
		S3DVertex ver[4];
		ver[0] = S3DVertex(-100, 0, 100,  0,0,0, 0x400000FF, 0, 0);
		ver[1] = S3DVertex(-100, 0,-100,  0,0,0, 0x400000FF, 0, 1);
		ver[2] = S3DVertex( 100, 0,-100,  0,0,0, 0x400000FF, 1, 0);
		ver[3] = S3DVertex( 100, 0, 100,  0,0,0, 0x400000FF, 1, 1);
		m_driver->drawIndexedTriangleList(ver, 4, index, 2);

		//ɖ߂
		m_driver->setMaterial(baseMaterial);
	}
};

#endif // INCLUDED_MDLVIEWER